/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

#include "conq.h"
#include "mplay.h"  /* flush this eventually */
#include "cconq.h"

#define request_input(x,y) beep();

/* (add a command to repeat the last command) */
/* (except for movement?) */
/* (ask_pos should echo distance continuously) */

char *ask_string();

Unit *
find_next_mover(side)
Side *side;
{
    Unit *nextunit, *unit;

    nextunit = find_next_actor(side, dui.curunit);
    if (nextunit != NULL) {
	make_current(nextunit);
    }
    return nextunit;
}

do_find_next_mover()
{
    Unit *oldcurunit = dui.curunit, *nextmover;

    nextmover = find_next_mover(dside);

    if (nextmover == NULL) {
	cmd_error("no units???");
    } else if (dui.curunit == oldcurunit) {
	notify("No other units to move!");
    }
}

do_find_prev_mover()
{
    Unit *nextunit;

    if ((nextunit = find_prev_actor(dside, dui.curunit)) != NULL) {
	make_current(nextunit);
    } else {
	beep();
    }
}

/* This is a clever (if I do say so myself) command to examine all occupants */
/* and suboccupants, in an inorder fashion. */

/* Should have an option to open up a list window that shows everything
   all at once. */

do_occupant(unit)
Unit *unit;
{
    Unit *nextup;

    if (unit->occupant != NULL) {
	make_current(unit->occupant);
    } else if (unit->nexthere != NULL) {
	make_current(unit->nexthere);
    } else {
	nextup = unit->transport;
	if (nextup != NULL) {
	    while (nextup->transport != NULL && nextup->nexthere == NULL) {
		nextup = nextup->transport;
	    }
	    if (nextup->nexthere != NULL)
	      make_current(nextup->nexthere);
	    if (nextup->transport == NULL)
	      make_current(nextup);
	} else {
	    /* This is a no-op if there is no stacking within a hex. */
	    make_current(unit_at(unit->x, unit->y));
	}
    }
}

/* Wake *everything* (that's ours) within the given radius.  Two commands */
/* actually; lowercase is transports only, uppercase is everybody. */

do_wakeup()
{
    wakeup_area(dside, dui.cmdarg, TRUE);
}

/* Wake up only the main/current unit in a hex. */

do_wakemain()
{
    wakeup_area(dside, dui.cmdarg, FALSE);
}

/* The area wakeup. */

int tmpflag;

wake_at(x, y)
int x, y;
{
    Unit *unit = unit_at(x, y);

    if (unit != NULL && (unit->side == tmpside || tmpside->designer)) {
/*	wake_unit(unit, tmpflag, WAKEFULL, (Unit *) NULL); */
	if (unit->plan && unit->act->acp > 0) unit->plan->asleep = FALSE;
    }
}

wakeup_area(side, n, occs)
Side *side;
int n, occs;
{
    tmpside = side;
    tmpflag = occs;
    apply_to_area(dui.curx, dui.cury, n, wake_at);
}

/* Put unit to sleep for a while. */

do_sentry(unit)
Unit *unit;
{
/*    order_sentry(unit, dui.cmdarg);  */
    if (unit->plan) {
	unit->plan->asleep = TRUE;
    }
}

do_sit(unit)
Unit *unit;
{
    if (unit->plan) {
	unit->plan->reserve = TRUE;
    }
}

/* Set unit to move to a given location.  Designers do a teleport. */

do_moveto(unit)
Unit *unit;
{
    int x, y;

    sprintf(spbuf, "Move %s to where?", unit_handle(dside, unit));
    if (ask_position(spbuf, &x, &y)) {
	if (dside->designer) {
	    leave_hex(unit);
	    enter_hex(unit, x, y);
	    make_current(unit);
	    all_see_hex(dui.curx, dui.cury);
	} else {
	    order_moveto(unit, x, y);
	}
    }
}

/* Determine how far away another point is.  */

do_distance()
{
    int x, y;

    if (ask_position("Distance to where?", &x, &y)) {
	notify("Distance is %d hexes.", distance(dui.curx, dui.cury, x, y));
    }
}

/* Auxiliary stuff used when searching for place to return to.  Note that */
/* a good refueling spot will be woken up, so it won't get too far away */
/* before unit has a chance to get there. */
/* Won't find refueling places inside other units, sigh. */

refuel_here(x, y)
int x, y;
{
    Unit *unit;

    for_all_stack(x, y, unit) {
	/* what about allies? */
	if (unit->side == tmpside && can_carry(unit, tmpunit)) {
	    /* this should be controlled by doctrine? */
	    /* shouldn't wake up, should get a new task to "wait up"
	       or even approach if possible */
	/*    wake_unit(unit, FALSE, WAKEOWNER, NULL);  */
	    return TRUE;
	}
	/* should look at occupants too */
    }
    return FALSE;
}

/* Search for a friendly refueler within range and set course for it.  */
/* Warn player and refuse to move if nothing close enough. */

/* (Should change to be appropriate for all unit types.) */
/* search should account for different possible paths... */

/* (move to kernel code, in unit planning) */

do_return(unit)
Unit *unit;
{
    int x, y, u = unit->type, r, range = area.maxdim;

    for_all_material_types(r) {
	if (um_tomove(u, r) > 0) {
	    range = min(range, unit->supply[r] / um_tomove(u, r));
	}
    }
    tmpside = dside;
    tmpunit = unit;
    if (refuel_here(unit->x, unit->y)) {
	cmd_error("Already at a resupply point!");
    } else if (search_area(unit->x, unit->y, range, refuel_here, &x, &y, 1)) {
	/* needs to be a task */
	/* should tweak plan so that most efficient route chosen */
	order_moveto(unit, x, y);
    } else {
	cmd_error("No resupply point in range!");
    }
}

/* Set orders to follow a leader unit. */

x_follow()
{
#if 0
    Unit *leader;

    if (grok_position()) {
	if ((leader = unit_at(dui.reqposx, dui.reqposy)) != NULL &&
	    leader->side == dside) {
	    if (leader != dui.requnit) {
		order_follow(dui.requnit, leader, dui.reqvalue2);
	    } else {
		cmd_error("Unit can't follow itself!");
	    }
	} else {
	    cmd_error("No unit to follow!");
	}
/*	restore_cur();  */
    } else {
	/* say something useful about curxy?? */
	request_input(x_follow, NULL);
    }
#endif
}

/* The command proper. */

do_fire(unit)
Unit *unit;
{
    int x, y;
    Unit *unit2;

    sprintf(spbuf, "Fire %s at where?", unit_handle(dside, unit));
    if (ask_position(spbuf, &x, &y)) {
	if ((unit2 = unit_at(x, y)) != NULL) {
	    prep_fire_at_action(unit, unit2, 0);
	} else {
	    beep();
	}
    }
}

do_follow()
{
#if 0
    ask_position("Which unit to follow?");
    dui.requnit = unit;
    request_input(x_follow, NULL);
#endif
}

/* Patrolling goes back and forth between two points.  First point is the */
/* current position. */

x_patrol()
{
#if 0
    if (grok_position()) {
	order_patrol(dui.requnit,
		     dui.requnit->x, dui.requnit->y,
		     dui.reqposx, dui.reqposy,
		     dui.reqvalue2);
	}
	restore_cur();
    } else {
	request_input(x_patrol, NULL);
    }
#endif
}

/* The command proper. */

do_patrol(unit)
Unit *unit;
{
#if 0
    ask_position("What other endpoint for patrol?");
    dui.requnit = unit;
    request_input(x_patrol, NULL);
/*    dui.reqposx = dui.front->curx;  dui.reqposy = dui.front->cury;  */
    dui.reqvalue2 = n;
#endif
}

/* Get rid of a unit. */

do_disband(unit)
Unit *unit;
{
    int u = unit->type;

    if (dside->designer) {
	kill_unit(unit, -1);
    } else if (u_hp_per_disband(u) > 0) {
	/* (text should come from game design) */
	notify("%s goes home.", unit_handle(dside, unit));
	if (do_disband_action(unit, unit)) {
	    notify("%s goes home.", unit_handle(dside, unit));
	} else {
	    /* failed, should say something */
	}
    } else {
	cmd_error("You can't just get rid of %s!",
		  unit_handle(dside, unit));
    }
}

/* Give a unit to another side (possibly to neutrals).  Units that won't */
/* change their sides when captured won't change voluntarily either. */

/* (but giving to neutral should be tested, otherwise might kill unit) */

do_give_unit(unit)
Unit *unit;
{
    int n = dui.cmdarg;
    int u = unit->type;

    if (/* u_change_side(u) || */ dside->designer) {
	unit_changes_side(unit, side_n(n), CAPTURE, PRISONER);
	all_see_hex(unit->x, unit->y);
	/* probably need to reset the unit's entire plan, including build */
   } else {
	cmd_error("You can't just give away the %s!", u_type_name(u));
    }
}

/* This can actually do general rearrangement, but defaults to putting the */
/* unit on the first available transport in the hex.  */
/* What about trying to embark a unit on itself or on its previous transp? */

do_embark(unit)
Unit *unit;
{
#if 0
    Unit *mainunit = unit_at(unit->x, unit->y);
    Unit *transport;

    if (mainunit != unit) {
#if 0
	    if (side->markunit == NULL ||
		side->markunit->x != unit->x || side->markunit->y != unit->y) {
		for_all_occupants(mainunit, transport) {
		    if (can_carry(transport, unit)) break;
		}
	    } else {
		transport = side->markunit;
	    }
#endif
	    if (transport != NULL) {
		/* needs some work... */
		transfer_unit_action(unit, unit, transport);
#if 0
		if (can_carry(transport, unit)) {
		    leave_hex(unit);
		    enter_transport(unit, transport);
		    if (n > 0) order_sentry(unit, n);
		    return TRUE;
		}
		else {
		    if (n > 0) order_embark(unit, n);
		    else cmd_error("Marked unit already full.");
		}
#endif
	    } else {
		if (n > 0) order_embark(unit, n);
		else cmd_error("No plausible transport!");
	    }
	} else {
	    cmd_error("Nothing for this unit to get into!");
	}
	return FALSE;
    }
#endif
}

/* Give supplies to a transport.  The argument tells how many to give. */

do_give(unit)
Unit *unit;
{
    int n = dui.cmdarg;
    int something = FALSE;
    int u = unit->type, m, r, gift, actual;
    Unit *main = unit->transport;

    if (main != NULL) {
	sprintf(spbuf, "");
	m = main->type;
	for_all_material_types(r) {
	    gift = (n < 0 ? (um_storage(m, r) - main->supply[r]) : n);
	    if (gift > 0) {
		something = TRUE;
		/* Be stingy if we're low */
		if (2 * unit->supply[r] < um_storage(u, r))
		    gift = max(1, gift/2);
		actual = transfer_supply(unit, main, r, gift);
		sprintf(tmpbuf, " %d %s", actual, m_type_name(r));
		strcat(spbuf, tmpbuf);
	    }
	}
	if (something) {
	    notify("%s gave%s.", unit_handle(dside, unit), spbuf);
	} else {
	    notify("%s gave nothing.", unit_handle(dside, unit));
	}
    } else {
	cmd_error("Can't transfer supplies here!");
    }
}

/* Take supplies from transport.  Both the transport must have something */
/* left. */

do_take(unit)
Unit *unit;
{
    int n = dui.cmdarg;
    int something = FALSE;
    int u = unit->type, m, r, need, actual;
    Unit *main = unit->transport;

    if (main != NULL) {
	sprintf(spbuf, "");
	m = main->type;
	for_all_material_types(r) {
	    need = (n < 0 ? um_storage(u, r) - unit->supply[r] : n);
	    if (need > 0) {
		something = TRUE;
		/* Be stingy if we're low */
		if (2 * main->supply[r] < um_storage(m, r))
		    need = max(1, need/2);
		actual = transfer_supply(main, unit, r, need);
		sprintf(tmpbuf, " %d %s", actual, m_type_name(r));
		strcat(spbuf, tmpbuf);
	    }
	}
	if (something) {
	    notify("%s got%s.", unit_handle(dside, unit), spbuf);
	} else {
	    notify("%s needed nothing.", unit_handle(dside, unit));
	}
    } else {
	cmd_error("Can't transfer supplies here!");
    }
}

/* Take us out of the game while letting everybody else */
/* continue on. */

do_resign()
{
    Side *side2 = NULL;

    if (side_lost(dside)) {
	/* Assume that a losing side just wants out, no need to confirm. */
	if (active_display(dside)) {
	    /* Is this sufficient to remove a player entirely? */
	    close_display(dside);
	}
    } else {
	if (ask_bool("Do you really want to give up?", FALSE)) {
	    if (numsides > 2) {
/*		ask_side("Who do you want to surrender to?", NULL); */
		resign_game(dside, side2);
	    } else {
		resign_game(dside, NULL);
	    }
	}
    }
}

/* Leave quickly when the boss walks by.  One person can kill a multi-player
   game, which isn't too great, but the alternatives are complicated.
   The stats file will be left behind, to foment argument about who would
   have won...

   This routine also includes a trapdoor for freezing/unfreezing
   machine players when designing. */

/* (Have an extra confirm for designers not to lose unsaved work??) */

/* The exit command sets up a confirmation. */

do_exit()
{
    if (side_lost(dside)) {
	if (0 /* numhumans > 1 */) {  /* should be "num active displays" */
	    /* At this point, killing the game is the same as resigning. */
	    do_resign(dui.cmdarg);
	} else {
	    exit_cconq(dside);
	}
    } else if (ask_bool("Do you really want to quit now?", FALSE)) {
	/* should force a loss first */
	exit_cconq(dside);
    }
}

/* Stuff game state into a file.  By default, it goes into the current
   directory.  If building a scenario, we can specify just which parts
   of the game state are to be written. */

do_save()
{
    char *rawcontents;
    Module *module;
    Obj *contents;
    Side *side2;

#ifdef DESIGNERS
    if (dside->designer) {
	if ((rawcontents = ask_string("Data to write?", "everything"))
	    != NULL) {
	    /* need to be able to get this name from somewhere */
	    module = create_game_module("random.scn");
	    /* need something better to turn contents into a Lisp object */
	    contents = intern_symbol(rawcontents);
	    /*	interpret_content_spec(module, contents);  */
	    notify("Module will be written to \"%s\" ...", module->filename);
	    if (write_game_module(module)) {
		notify("Done writing to \"%s\".", module->filename);
	    } else {
		cmd_error("Can't open file \"%s\"!", module->filename);
	    }
	    return;
	} else {
	    return;
	}
    }
#endif /* DESIGNERS */
    if (0 /* checkpointing not allowed */) {
	if (ask_bool("You really want to save and exit?", FALSE)) {
	    notify_all("Game will be saved to \"%s\" ...", savefile_name());
	    if (write_entire_game_state(savefile_name())) {
		close_displays();
		/* this should be conditional? */
		exit(0);
	    } else {
		cmd_error("Can't open file \"%s\"!", savefile_name());
	    }
	}
    } else {
	if (write_entire_game_state(savefile_name())) {
	    notify_all("Game saved.");
	} else {
	    cmd_error("Can't open savefile \"%s\"!", savefile_name());
	}	    
    }
}

/* this should go somewhere */
#if 0 /* def PREVVIEW */
    if (n > 0) {  /* remove the views of units likely to have moved. */
	int x, y, ts;

      for_all_hexes(x, y)
	  if (side_view_age(x, y) > n) {
	    int viewunit = vtype(side_view(side, x, y));

	    if (viewunit != EMPTY && viewunit != UNSEEN
		&& mobile(viewunit)) { 
	      ts = side_view_timestamp(side, x, y);
	      set_side_view(side, x, y, EMPTY);
	      side_view_timestamp(side, x, y) = ts;
	    }
	  }
    }
#endif

/* Name or rename the current unit or a given side. */

do_name()
{
    char *newname;
    Side *side2;

    if (dui.curunit != NULL) {
	if ((newname = ask_string("New name for unit:", dui.curunit->name))
	    != NULL) {
	    if (strlen(newname) == 0) newname = NULL;
	    dui.curunit->name = newname;
	    update_unit_display(dside, dui.curunit);
	}
    } else {
	if ((newname = ask_string("New name for yourself:", dside->name))
	    != NULL) {
	    if (strlen(newname) == 0) newname = NULL;
	    dside->name = newname;
	    for_all_sides(side2) show_side_list(side2);
	}
    }
}

/* Move the current location as close to the center of the display as
   possible, and redraw everything. */

do_recenter()
{
    dui.vcx = dui.curx;  dui.vcy = dui.cury;
    set_map_viewport();
    redraw();
}

/* Redraw everything using the same code as when windows need a redraw. */

do_redraw()
{
    recalc_caches();
    redraw();
}

recalc_caches()
{
#if 0
    int u;
    Unit *unit;
    Side *side;

    for_all_sides(side) {
	for_all_unit_types(u) {
	    side->units[u] = 0;
	    side->unitsbuilding[u] = 0;
	}
	for_all_side_units(unit) {
	    if (alive(unit)) {
		if (completed(unit)) {
		    ++(side->units[unit->type]);
		} else {
		    ++(side->unitsbuilding[unit->type]);
		}
	    }
	}
    }
#endif
}

/* Stop the clock. */

/* (Needs some restrictions to prevent abuse.) */

do_pause()
{
    extern int paused;
    Side *side2;

    paused = !paused;
    for_all_sides(side2) {
	show_game_date(side2);
    }
}

/* The option command. */

do_options()
{
    char *xcmd;

    if ((xcmd = ask_string("ext cmd # ", NULL)) != NULL) {
	if (strcmp(xcmd, "print") == 0) {
	    dump_view();
	} else if (strcmp(xcmd, "?") == 0) {
	    notify("print,design,debug,debugm,recalc");
	} else if (strcmp(xcmd, "people") == 0) {
	    dui.drawpeople = !dui.drawpeople;
	    redraw();
	} else
#ifdef DESIGNERS
	if (strcmp(xcmd, "design") == 0) {
	    /* should check password eventually? */
	    if (dside->designer) {
		become_nondesigner(dside);
	    } else {
		become_designer(dside);
	    }
	} else
#endif /* DESIGNERS */
#ifdef DEBUGGING
	if (strcmp(xcmd, "debug") == 0) {
#ifndef Debug
	    Debug = !Debug;
#endif
	} else if (strcmp(xcmd, "debugm") == 0) {
#ifndef DebugM
	    DebugM = !DebugM;
#endif
	} else if (strcmp(xcmd, "recalc") == 0) {
	    doublecheck_state();
	    reset_coverage();
	    redraw();
	} else
#endif /* DEBUGGING */
	{
	    cmd_error("unrecognized ext cmd \"%s\"", xcmd);
	}
    }
}

do_build(unit)
Unit *unit;
{
    int u2;

    if (!can_build(unit)) {
	cmd_error("%s can't build anything!", unit_handle(dside, unit));
#if 0
    } else if (!u_occproduce(unit->type) && unit->transport != NULL) {
	cmd_error("%s can't build anything while inside another unit!",
		  unit_handle(dside, unit));
#endif
    } else {
	if ((u2 = ask_unit_type("Type to create:", NULL)) > 0) {
	    notify(dside, "%s will build %s",
		   unit_handle(dside, unit), u_type_name(u2));
	    push_build_task(unit, u2);
	} else {
	    /* should clear toplines */
	}
    }
}

/* Setting what a unit will produce next. */

do_next_build(unit)
Unit *unit;
{
}

/* Set a unit to not produce anything (yes, this really is useful). */

do_idle(unit)
Unit *unit;
{
    int n = dui.cmdarg;

    /* (blast task and/or plan) */
}

/* Send a short (1 line) message to another player.  Some messages are */
/* recognized specially, causing various actions. */

do_message()
{
    char *msg;
    Side *side3 = NULL;

    if ((msg = ask_string("Message: ", "")) != NULL) {
	if (strlen(msg) == 0) {
	    notify("You keep your mouth shut.");
	} else if (1 /* broadcast? */) {
	    if (strlen(msg) > 0) {
		notify_all("The %s announce: %s",
			   plural_form(dside->name), msg);
	    }
#if 0
	} else if (strcmp(msg, "briefing") == 0) {
	    notify(dui.reqside, "Receiving a briefing from the %s...",
		   plural_form(side->name));
	    reveal_side(side, dui.reqside, 100);
	    notify("You just briefed the %s on your position.",
		   plural_form(dui.reqside->name));
	} else if (strcmp(msg, "alliance") == 0) {
	    notify("You propose a formal alliance with the %s.",
		   plural_form(dui.reqside->name));
	    side->relationships[side_number(dui.reqside)] = ALLY;
	    if (dui.reqside->relationships[side_number(side)] >= ALLY) {
		declare_alliance(side, dui.reqside);
		for_all_sides(side3) redraw(side3);
	    } else {
		notify(dui.reqside, "The %s propose a formal alliance.",
		       plural_form(side->name));
	    }
	} else if (strcmp(msg, "neutral") == 0) {
	    notify("You propose neutrality with the %s.",
		   plural_form(dui.reqside->name));
	    side->relationships[side_number(dui.reqside)] = NEUTRAL;
	    if (dui.reqside->relationships[side_number(side)] == NEUTRAL) {
		declare_neutrality(side, dui.reqside);
		for_all_sides(side3) redraw(side3);
	    } else {
		notify(dui.reqside, "The %s propose neutrality.",
		       plural_form(side->name));
	    }
	} else if (strcmp(msg, "war") == 0) {
	    notify("You declare war on the %s!",
		   plural_form(dui.reqside->name));
	    declare_war(side, dui.reqside);
	    for_all_sides(side3) redraw(side3);
#endif
	} else {
	    notify(side3, "The %s say to you: %s",
		   plural_form(dside->name), msg);
	}
    }
}

/* Toggle the followaction flag. */

do_look()
{
    dui.followaction = !dui.followaction;
    if (dui.followaction) {
	notify("Following the action.");
    } else {
	notify("Not following the action.");
    }
}

/* Add a new player to the game. */
/* Should use arg to decide whether to convert machine player (or just */
/* use if available?)  Also needs to decide if new player is human and */
/* which host to open, then go through side's startup seq and open disp. */
/* Issues of cached values (?) and war/alliance setup and uses of numsides. */

do_add_player()
{
    notify("Sorry, can't add new players yet!");
}

/* Bring up help info. */

do_help()
{
    /* Switch screen to displaying help info. */
    dui.screenmode = helpwin;
    /* Switch input handling to interpret help commands. */
    dui.mode = HELP;
    /* Get a help node and display its content. */
    if (dui.curhelpnode == NULL) {
	dui.curhelpnode = firsthelpnode;
    }
    dui.helpstring = get_help_text(dui.curhelpnode);
    show_help();
    refresh();
}

/* Display the program version. */

do_version()
{
    notify("Curses Xconq version %s", version_string());
    notify("(c) %s", copyright_string());
}

/* Set the unit to automatic control.  */

do_auto(unit)
Unit *unit;
{
    if (unit->plan) {
	unit->plan->autotask = !unit->plan->autotask;
	/* finish move under automatic control */
    }
}

/* Predeclarations of all commands. */

int
do_change_list_order()
{
    ++(dui.listtype);
    dui.listtype %= numlisttypes;
    show_list();
    refresh();
}

int
do_change_list_sides()
{
    ++(dui.listsides);
    dui.listsides %= numlistsides;
    show_list();
    refresh();
}

/* Command to declare that we've finished moving for this turn. */

do_finished()
{
    finish_turn(dside);
}

Unit *lastactor = NULL;

/* Command to toggle between interaction modes. */

do_survey_mode()
{
    if (dui.mode == MOVE) {
	dui.mode = SURVEY;
    } else {
	dui.mode = MOVE;
	make_current(find_next_actor(dside, lastactor));
    }
    show_map();
}

/* Commands to change the dividing line between the right-hand and left-hand
   windows of the screen. */

do_grow_map()
{
    resize_map(dui.cmdarg);
}

do_shrink_map()
{
    resize_map(0 - dui.cmdarg);
}

resize_map(n)
int n;
{
    /* Resize the left-hand-side windows. */
    dui.mw += n;
    dui.ww[dui.closeup] += n;
    dui.ww[dui.map] += n;
    /* Move and resize the right-hand-side windows. */
    dui.lw -= n;
    dui.wx[dui.sides] += n;
    dui.ww[dui.sides] -= n;
    dui.wx[dui.list] += n;
    dui.ww[dui.list] -= n;
    /* Update the screen to refect the changes. */
    set_scroll();
    redraw();
}

#ifdef DESIGNERS

/* Commands that designers use to build stuff. */

/* Create a unit of given type, position, and side. */

x_add_unit()
{
    int u;
    Unit *unit;

    if ((u = grok_unit_type()) != NONUTYPE) {
	if ((unit = create_unit(u)) != NULL) {
	    set_unit_side(unit, dside);
	    enter_hex(unit, dui.curx, dui.cury);
	    init_supply(unit);
	    all_see_hex(dui.curx, dui.cury);
	    make_current(unit);
	} else {
	    cmd_error("Unit creation failed!");
	}
    }
}

/* The command proper just asks for the type of unit to create. */

do_add_unit()
{
    if (dside->designer) {	
	ask_unit_type("Type of unit to create: ", NULL);
	request_input(x_add_unit, NULL);
    } else {
	cmd_error("No permission to create units!");
    }
}

/* Function to change just one hex and to echo that change. */

set_one_hex(x, y)
int x, y;
{
    set_terrain_at(x, y, tmpttype);
    see_exact(tmpside, x, y);
    update_cell_display(tmpside, x, y, FALSE);
}

/* Basic terrain painting command. */

x_paint_terrain()
{
    int t;

    if ((t = grok_terrain_type()) != NONTTYPE) {
	tmpside = dside;
	tmpttype = t;
	apply_to_hexagon(dui.curx, dui.cury, 1, 1, set_one_hex);
    } else {
	/* an error, ask again */
	request_input(x_paint_terrain, NULL);
    }
}

do_paint_terrain()
{
    if (dside->designer) {	
	ask_terrain_type("Type of terrain to paint: ", NULL);
	request_input(x_paint_terrain, NULL);
    } else {
	cmd_error("No permission to paint terrain!");
    }
}

#endif /* DESIGNERS */

/* The following structure is used only for the dispatch table. */

typedef struct func_tab {
    char fchar;                 /* character to match against */
    int (*code)();              /* pointer to command's function */
    int defaultarg;             /* default for argument */
    int needunit;              /* true if cmd always applies to a unit */
    char *help;                 /* short documentation string */
} FuncTab;

/* The command table itself.  Default iterations of ITERFLAG actually turn */
/* into the value of "itertime", which can be set from doctrine. */

FuncTab commands[] = {
    's', do_sentry, ITERFLAG, TRUE, "put a unit on sentry duty",
    'w', do_wakemain, 0, FALSE, "wake units up from whatever they were doing",
    ' ', do_sit, 1, TRUE, "(Space Bar) make unit be on sentry this turn only",
    'r', do_return, ITERFLAG, TRUE, "return unit to nearest city/transport",
    'd', do_find_next_mover, 1, TRUE, "delay unit's move until later in turn",
    'f', do_fire, 1, TRUE, "fire at a designated location",
    'm', do_moveto, 1, TRUE, "move to given location",
/*    '\015', do_movetounit, ITERFLAG, TRUE, "move to the marked unit", */
    'z', do_survey_mode, 1, FALSE, "toggle between survey and move modes",
    'i', do_occupant, 1, TRUE, "look at the occupants",
    'e', do_embark, 0, TRUE, "embark units onto transport occupying same unit",
/*    '\006', do_fill, ITERFLAG, TRUE, "order transport to sleep until full", */
    'o', do_options, 1, FALSE, "set options",
    'g', do_give, -1, TRUE, "give supplies to the transport",
    't', do_take, -1, TRUE, "take supplies from the transport",
/*    'c', do_center, 1, FALSE, "designate current location as center of action",  */
    '\r', do_finished, 1, FALSE, "declare turn to be finished",
    ' ', NULL, 0, 0, "",
    'P', do_build, 1, TRUE, "set/change unit construction",
    'W', do_wakeup, 0, FALSE, "wake ALL units in area, including occupants",
    'D', do_disband, 1, TRUE, "disband a unit and send it home",
/*    'E', do_movetotransport, ITERFLAG, TRUE, "move to filling transport", */
    'p', do_patrol, ITERFLAG, TRUE, "go on a patrol",
    'G', do_give_unit, 0, TRUE, "turn unit over to another side",
    'M', do_message, 0, FALSE, "send a message to other sides",
    'X', do_resign, 0, FALSE, "resign from the game",
    'Q', do_exit, 1, FALSE, "kill game for all players",
    'S', do_save, 1, FALSE, "save the game into a file",
/*    'A', do_add_player, 1, FALSE, "add a new player to the game", */
    'C', do_name, 1, FALSE, "call a unit or side by a name",
    '>', do_grow_map, 5, FALSE, "increase map area",
    '<', do_shrink_map, 5, FALSE, "decrease map area",
    '?', do_help, 1, FALSE, "display help info",
    'V', do_version, 1, FALSE, "display program version",
    'A', do_auto, 1, TRUE, "Turn unit over to machine control",
    '\001', do_auto, 1, TRUE, "Turn unit over to machine control",
    '\020', do_next_build, 1, TRUE, "set next unit type to construct",
    '\014', do_recenter, 1, FALSE, "center and redraw around cursor",
    '\022', do_redraw, 10, FALSE, "redraw screen only",
    '\006', do_look, -1, FALSE, "look at action point when not your turn",
    '#', do_distance, 1, FALSE, "measure the distance from the current point",
#ifdef DESIGNERS
    '\\', do_add_unit, 1, FALSE, "build a new unit (designers only)",
    '`', do_paint_terrain, 1, FALSE, "modify terrain (designers only)",
#endif /* DESIGNERS */
    '\0', NULL, 0, FALSE, NULL   /* end of table marker */
};

FuncTab mapcommands[] = {
    '\0', NULL, 0, FALSE, NULL   /* end of table marker */
};

FuncTab listcommands[] = {
    '+', do_change_list_order, -1, FALSE, "change list sort order",
    '_', do_change_list_sides, -1, FALSE, "change list side selection",
    '\0', NULL, 0, FALSE, NULL   /* end of table marker */
};

/* Search in command table and execute function if found, complaining if
   the command is not recognized.  Many commands operate on the "current
   unit", and all uniformly error out if there is no current unit, so put
   that test here.  Also fix up the arg if the value passed is one of the
   specially recognized ones. */

execute_command()
{
    struct func_tab *cmd;
    
    for (cmd = commands; cmd->fchar != '\0'; ++cmd) {
	if (dui.inpch == cmd->fchar) {
	    if (dui.cmdarg == DFLTFLAG) dui.cmdarg = cmd->defaultarg;
	    if (dui.cmdarg == ITERFLAG) dui.cmdarg = dside->itertime;
	    DGprintf("... actual arg is %d\n", dui.cmdarg);
	    if (cmd->needunit) {
		if (dui.curunit == NULL) {
		    cmd_error("Nothing to command here!");
		} else if (!side_controls_unit(dside, dui.curunit)) {
		    cmd_error("You don't control the unit!");
		} else {
		    (*(cmd->code))(dui.curunit);
		}
	    } else {
		(*(cmd->code))();
	    }
	    /* map- or list-specific commands here */
	    return;
	}
    }
    cmd_error("Unknown command '%c'", dui.inpch);
}

/* Paste one-line descriptions of commands into the supplied buffer. */

describe_commands(arg, key, buf)
int arg;
char *key, *buf;
{
    FuncTab *cmd;

    sprintf(spbuf, "To move a unit, use [hlyubn]\n");
    append_to_buffer(buf, spbuf);
    sprintf(spbuf, "To select a unit, use survey mode\n");
    append_to_buffer(buf, spbuf);
    sprintf(spbuf, "[HLYUBN] moves unit repeatedly in that direction\n");
    append_to_buffer(buf, spbuf);
    sprintf(spbuf, "\n");
    append_to_buffer(buf, spbuf);
    for (cmd = commands; cmd->fchar != '\0'; ++cmd) {
	if (cmd->fchar < ' ' || cmd->fchar > '~') { 
	    sprintf(spbuf, "^%c", (cmd->fchar^0x40));
	    append_to_buffer(buf, spbuf);
	} else {
	    sprintf(spbuf, "%c ", cmd->fchar);
	    append_to_buffer(buf, spbuf);
	}
	sprintf(spbuf, " %s\n", cmd->help);
	append_to_buffer(buf, spbuf);
    }
}

/* Generic command error routine just does a notify. */

/*VARARGS*/
cmd_error(control, a1, a2, a3, a4, a5, a6)
char *control, *a1, *a2, *a3, *a4, *a5, *a6;
{
    notify(control, a1, a2, a3, a4, a5, a6);
    beep();
}
