(game-module "greek"
  (blurb "Ancient Greece ca 350 BC")
  (variants (world-seen) (world-size))
)

(set see-all true)

(unit-type peltast
  (help "light fast infantry"))
(unit-type hoplite
  (help "heavy destructive infantry"))
(unit-type archer
  (help "archers and slingers"))
(unit-type cavalry
  (help "more like light cavalry"))
(unit-type trireme
  (help "three-decked ship - mainstay of the navies"))
(unit-type elephant
  (help "the tanks of the ancient world"))
(unit-type |siege engine|
  (help "bash cities"))
(unit-type camp
  (help "quickly thrown up, not too useful"))
(unit-type polis
  (help "typical city-state"))
(unit-type metropolis
  (help "a large and powerful city"))

(define p peltast)
(define h hoplite)
(define a archer)
(define c cavalry)
(define T trireme)
(define E elephant)
(define S |siege engine|)
(define / camp)
(define * polis)
(define @ metropolis)

(add S image-name "catapult")
(add / image-name "camp")
(add * image-name "walltown")
(add @ image-name "parthenon")

(material-type food
  (help "everybody needs food to survive"))
(material-type people)

(terrain-type sea (color "sky blue") (char "."))
(terrain-type plains (color "green") (char "+"))
(terrain-type forest (color "forest green") (char "%"))
(terrain-type desert (color "yellow") (char "~"))
(terrain-type mountains (color "sienna") (char "^"))
(terrain-type river (color "blue") (char "-")
  (subtype border))

(define cities (/ * @))
(define makers (* @))
(define movers (p h a c T E S))
(define water (sea river))
(define ship (T))
(define land (plains forest desert mountains))
(define land-u (p h a c E S))

;;; Static relationships.

(table vanishes-on
  (land-u water 10)
  (cities water 10)
  (ship land 10)
  )

(add cities capacity 10)

(add trireme capacity 6)

(table unit-size-as-occupant
  ((p h a c) trireme (2 2 1 3))
  )

;;; Actions.

(add u* acp-per-turn  (2 1 1 2 2 1 1 0 1 1))

;;; Movement.

(add movers speed (2 1 1 2 2 1 1))

(table mp-to-enter-terrain
  (land-u water 10)
  (ship land 10)
  )

;true [ * @ ] always-seen

;;; Construction.

;(add u* cp (4 8 10 10 20 40))

(table acp-to-create
  (* movers 1)
  (@ movers 1)
  )

(table cp-on-creation
  (* movers 1)
  (@ movers 1)
  )

(table acp-to-build
  )

(table cp-per-build
  )

;;; Combat.

(add u* hp-max (1 1 1 1 3 5 2 10 20 40))

(table hit-chance
  (p u* (40 30 50 40 10 10 30 30 20 10))
  (h u* (50 40 50 50 20 20 50 70 20 10))
  (a u* (30 20 30 30 20 60 30  0  0  0))
  (c u* (60 30 50 50 20 50 60 50  0  0))
  (T u* (10 10 10 10 50 10 30  0  0  0))
  (E u* (50 50 50 50 10 50 20 40  0  0))
  (S u* (10 10 10 10 10 10  5 99 99 99))
  (/ u* ( 0  0  0  0  0 20  0  0  0  0))
  (* u* (40 30 30 40  0 10 20  0  0  0))
  (@ u* (40 30 30 40  0 10 20  0  0  0))
  )

(table damage
  (u* u* 1)
  ((h E) u* 2)
  (S @ 4)
  )

(table capture-chance
  (h * 10)
  (h @ 5)
  )

;; 1000 c attrition

;;; Random generation.

(set alt-blob-density 10000)
(set alt-blob-size 20)
(set alt-smoothing 6)
(set wet-blob-density 2000)
(set wet-blob-size 100)

(add t* alt-percentile-min (  0  70  70  70  90 0))
(add t* alt-percentile-max ( 70  90  90  90 100 0))
(add t* wet-percentile-min (  0  20  80   0   0 0))
(add t* wet-percentile-max (100  80 100  20 100 0))

(set edge-terrain sea) ; the river Oceanus

(add @ start-with 1)
(add h start-with 1)
; Try to get countries on the coast.
(add (sea plains) country-terrain-min (1 1))

(table independent-density
  (* (plains desert) (100 50)))

(table favored-terrain
  (u* t* 100)
  (cities sea 0)
  )

;;; (need to set up a BC calendar)

(world 2500 (year-length 12))  ; big world, can't circumnavigate.

;;; Side names.  The more famous ones are weighted more heavily.

(side-library set
  (10 (name "Athens") (adjective "Athenian"))
  (10 (name "Sparta") (adjective "Spartan"))
  (5 (name "Corinth") (adjective "Corinthian"))
  (5 (name "Thebes") (adjective "Theban"))
  (3 (name "Argos") (adjective "Argive"))
  (2 (name "Megara") (adjective "Megaran"))
  (2 (name "Miletus") (adjective "Miletan"))
  (2 (name "Messene") (adjective "Messenian"))
  (2 (name "Syracuse") (adjective "Syracusan"))
  (2 (name "Ephesus") (adjective "Ephesian"))
  (2 (name "Delos") (adjective "Delian"))
  ((name "Lemnos") (adjective "Lemnian"))
  ((name "Ambracia") (adjective "Ambraciot"))
  ((name "Phokia") (adjective "Phokian"))
  ((name "Chios") (adjective "Chian"))
  ((name "Gelos") (adjective "Geloan"))
  ((name "Caria") (adjective "Carian"))
  ((name "Lokria") (adjective "Lokrian"))
  ((name "Melos") (adjective "Melian"))
  ((name "Phlias") (adjective "Phliasian"))
  ((name "Samos") (adjective "Samian"))
  ((name "Thuria") (adjective "Thurian"))
  ((name "Tegea") (adjective "Tegean"))
  ((name "Dolope") (adjective "Dolopian"))
  ((name "Olynthia") (adjective "Olynthian"))
  ((name "Elis") (adjective "Elean"))
  ((name "Lucania") (adjective "Lucanian"))
  ((name "Kythera") (adjective "Kytheran"))
  ((name "Skios") (adjective "Skionian"))
)

(game-module (notes
  "This time strictly covers about 600 to 200 B.C., since later times were"
  "dominated by Romans with rather different sorts of units.  Land warfare"
  "was most significant, with some notable sea-fights.  Cities were nearly"
  "untakeable, so the action centered around sieges and field battles."
  ""
  "The numbers border on the plausible, but again this one has not been"
  "played enough to find the imbalances even, let alone decide on good"
  "strategies."

  "(following should be individual unit notes)"
  "Peltast (light infantry).  Useful in skirmishes."
  ""
  "Hoplite (heavy infantry).  For the capture of cities by assault."
  ""
  "Archer."
  ""
  "Cavalry.  Fast, but ineffective against cities."
  ""
  "Trireme (larger ship)."
  ""
  "Elephant.  Can destroy cavalry, but vulnerable to archers, who shoot"
  "arrows into the elephants' eyes and panic them."
  ""
  "Siege Engine.  Bash cities with this."
))
