(game-module "panzer"
  (title "Panzer")
  (blurb "WW II, the Eastern Front")
  (variants (see-all true) (world-size 80 40))
)

(terrain-type water (color "sky blue") (image-name "water") (char ".")
  )
(terrain-type swamp (color "yellow green") (char "=")
  )
(terrain-type gully (color "sienna") (char "&")  ; require explicit icon?
  )
(terrain-type woods (color "green") (image-name "forest") (char "%")
  )
(terrain-type slope (color "goldenrod") (char "/")
  )
(terrain-type clear (color "khaki") (char "+")
  )
(terrain-type thick-forest (color "forest green") (image-name "forest")
  (subtype border)
  )
(terrain-type stream (color "blue") (image-name "water")
  (subtype connection)
  )
(terrain-type road (color "gray")
  (subtype connection)
  )

;;; Ground is what doesn't require swimming.

(define ground (swamp gully clear woods slope))

;;; Vehicles require solid ground.

(define solid (gully clear woods slope))

;;; Wheeled vehicles can't deal with woods.

(define drivable (gully clear slope))

(add t* elevation-min 0)
(add t* elevation-max 150)

;;; Definitions of all the unit types.

;;; Random gen should do these in clumps?
(unit-type town (image-name "town20"))
(unit-type block)
(unit-type minefield)
(unit-type cmdpost (name "command post") (image-name "cmd-post")
  (acp-per-turn 1) (speed 1))

;;; Russian unit types.

(unit-type rmg (name "MG")
  (help "12.7mm machine gun"))
(unit-type r45 (name "45mm AT"))
(unit-type r57 (name "57mm AT"))
(unit-type r76at (name "76.2mm AT"))
(unit-type r76h (name "76.2mm H"))
(unit-type r122 (name "122mm AT"))
(unit-type r82m (name "82mm M (Mot)")
  (help "Russian motorized 82mm mortar"))
(unit-type r82 (name "82mm M")
  (help "Russian 82mm mortar"))
(unit-type r120 (name "120mm M")
  (help "Russian 120mm mortar - a powerful long-range hitter"))

(define r-arty (rmg r45 r57 r76at r76h r122 r82m r82 r120))

(add (r45 r57 r76at r122) image-name "pz-at")
(add rmg image-name "pz-flak")
(add r76h image-name "pz-how")
(add (r82 r82m r120) image-name "pz-mortar")

(add r-arty possible-sides "russian")
(add r-arty acp-per-turn 1)
(add r-arty speed 0)
(add (r82m r82) speed 2)

(add r-arty acp-to-fire 1)

(add r-arty range ( 6  3  4  5 10 20 12 12 20))

(unit-type reng (name "engineers(r)") (image-name "engineers"))
(unit-type rrecon (name "recon(r)") (image-name "infantry"))
(unit-type rrifle (name "rifle(r)") (image-name "infantry"))
(unit-type rguard (name "guards(r)") (image-name "infantry"))
(unit-type rsmg (name "submachinegun(r)") (image-name "infantry"))

(define r-inf (reng rrecon rrifle rguard rsmg))

(add r-inf possible-sides "russian")
(add r-inf acp-per-turn 2)
(add r-inf speed 2)

(unit-type rcav (name "cavalry") (image-name "cavalry")
  (acp-per-turn 6) (speed 6))

(add rcav possible-sides "russian")

(unit-type rwag (name "wagon(r)"))
(unit-type rtruck (name "truck(r)"))
(unit-type rhalf (name "halftrack(r)"))

(define r-transp (rwag rtruck rhalf))

(add r-transp possible-sides "russian")

(add r-transp acp-per-turn (6 24 20))
(add r-transp speed (6 24 20))

(unit-type su-152 (name "SU-152") (image-name "su-152"))
(unit-type su-76 (name "SU-76") (image-name "su-76"))
(unit-type su-85 (name "SU-85") (image-name "su-85"))
(unit-type su-100 (name "SU-100") (image-name "su-100"))
(unit-type jsu-122 (name "JSU-122") (image-name "jsu-122"))

(define r-td (su-152 su-76 su-85 su-100 jsu-122))

(add r-td possible-sides "russian")

(add r-td acp-per-turn (14 18 22 16 14))
(add r-td speed    (14 18 22 16 14))

(add r-td acp-to-fire  ( 7  9 11  8  7))
(add r-td range    (10  5  8  8 10))

(unit-type kv85 (name "KV85") (image-name "kv85"))
(unit-type t34 (name "T-34c") (image-name "t-34c")
  (help "The most common type of Russian tank"))
(unit-type t34/85 (name "T-34/85") (image-name "t-34-85"))
(unit-type js2 (name "JS II") (image-name "js-2"))
(unit-type js3 (name "JS III") (image-name "js-3"))

(define r-tank (kv85 t34 t34/85 js2 js3))

(add r-tank possible-sides "russian")

(add r-tank acp-per-turn (20 22 22 16 16))
(add r-tank speed    (20 22 22 16 16))

(add r-tank acp-to-fire  (10  5 11  8  8))
(add r-tank range    ( 8  6  8  8 12))

; (should be formed by appending)
;(define russian (r-arty r-inf rcp rcav r-transp r-tank))

;;; German unit types.

(unit-type g50 (name "50mm AT"))
(unit-type g75 (name "75mm AT"))
(unit-type g88 (name "88mm AT"))
(unit-type g20 (name "20mm F") (help "German 20mm flak"))
(unit-type g20q (name "20mmQUAD") (help "German 20mm quad flak"))
(unit-type g75h (name "75mm H"))
(unit-type g150 (name "150mm H")
  (help "German 150mm howitzer - hard-hitting but shortish range"))
(unit-type g81 (name "81mm M"))
(unit-type g120 (name "120mm M(g)"))

(define g-arty (g50 g75 g88 g20 g20q g75h g150 g81 g120))

(add (g50 g75 g88) image-name "pz-at")
(add (g75h g150) image-name "pz-how")
(add (g20 g20q) image-name "pz-flak")
(add (g81 g120) image-name "pz-mortar")

(add g-arty possible-sides "german")
(add g-arty acp-per-turn 1)
(add g-arty speed 0)

(add g-arty acp-to-fire 1)
(add g-arty range ( 5  6 20 10 10 12 12 12 20))

(unit-type geng (name "engineers(g)") (image-name "engineers"))
(unit-type gsec (name "security(g)") (image-name "infantry"))
(unit-type grifle (name "rifle(g)") (image-name "infantry"))
(unit-type gsmg (name "submachinegun(g)") (image-name "infantry"))

(define g-inf (geng gsec grifle gsmg))

(add g-inf possible-sides "german")
(add g-inf acp-per-turn 2)
(add g-inf speed 2)

(unit-type gwag (name "wagon(g)"))
(unit-type gtruck (name "truck(g)"))
(unit-type ghalf (name "halftrack(g)"))

(define g-transp (gwag gtruck ghalf))

(add (rwag gwag) image-name "wagon")
(add (rtruck gtruck) image-name "truck")
(add (rhalf ghalf) image-name "halftrack")

(add g-transp possible-sides "german")

(add g-transp acp-per-turn (6 24 20))
(add g-transp speed (6 24 20))

(unit-type sdkfz/1 (name "SdKfz 234/1") (image-name "sdkfz-234-1"))
(unit-type puma (name "Puma"))
(unit-type sdkfz/4 (name "SdKfz 234/4") (image-name "sdkfz-234-4"))

(define g-car (sdkfz/1 puma sdkfz/4))

(add g-car possible-sides "german")

(add g-car acp-per-turn (32 28 28))
(add g-car speed        (32 28 28))

(add g-car acp-to-fire  (16 14 14))
(add g-car range        ( 2  6 13))

(unit-type maultier (name "Maultier")
  (help "Second-deadliest in the game, but very vulnerable"))
(unit-type wespe (name "Wespe")
  (help "Longest range in the game, but vulnerable"))
(unit-type hummel (name "Hummel")
  (help "Deadliest unit in the game, but vulnerable"))

(define g-spa (maultier wespe hummel))

(add g-spa possible-sides "german")

(add g-spa acp-per-turn (20 16 16))
(add g-spa speed        (20 16 16))

(add g-spa acp-to-fire  (10  8  8))
(add g-spa range        (12 32 24))

(define g-misc (puma sdkfz/4 sdkfz/1 maultier wespe hummel))

(unit-type gw38 (name "Gw 38"))
(unit-type wirbelwind (name "Wirbelwind"))
(unit-type stuh42 (name "Stu H 42"))
(unit-type marder3 (name "Marder III"))
(unit-type stug3 (name "StuG III"))
(unit-type hetzer (name "Hetzer"))
(unit-type nashorn (name "Nashorn"))
(unit-type jgdpz4 (name "Jgd Pz IV"))
(unit-type jgdpz5 (name "Jgd Pz V"))
(unit-type jgdpz6 (name "Jgd Pz VI"))

(define ww wirbelwind)

(define g-td (gw38 ww stuh42 marder3 stug3 hetzer nashorn jgdpz4 jgdpz5 jgdpz6))

(add g-td possible-sides "german")

(add g-td acp-per-turn (12 16 16 16 16 12 16 16 18 10))
(add g-td speed    (12 16 16 16 16 12 16 16 18 10))

(add g-td acp-to-fire  ( 6  8  8  8  8  6  8  8  9  5))
(add g-td range    (12 10 12  8  8  8 20 12 12 12))

(unit-type lynx (name "Lynx"))
(unit-type pz4 (name "PzKpfw IV") (image-name "pz-4")
  (help "Most common type of Panzer"))
(unit-type panther (name "Panther")
  (help "All-around best tank"))
(unit-type tiger1 (name "Tiger I") (image-name "tiger-1")
  (help "Powerful but slow tank"))
(unit-type tiger2 (name "Tiger II") (image-name "tiger-2"))

(define g-tank (lynx pz4 panther tiger1 tiger2))

(add g-tank possible-sides "german")

(add g-tank acp-per-turn (20 16 20 16 12))
(add g-tank speed    (20 16 20 16 12))

(add g-tank acp-to-fire  (10  8 10  8  6))

(add g-tank range    ( 4  8 12 10 12))

;(define german (append g-arty g-inf gcp g-transp g-car g-spa g-td g-tank))

;;; What happens to tanks that get wrecked.

(unit-type wreck (acp-per-turn 0))

;;; Wrecks don't belong to either side.

(add wreck possible-sides "independent")

;; This should be the end of type definitions.

;;; Static relationships.

(table vanishes-on
  (u* water true)
  (r-td swamp true)
  (r-tank swamp true)
  (g-misc swamp true)
  (g-td swamp true)
  (g-tank swamp true)
)

;; Russian units are larger than German units, so we can only stack two in a hex
;; instead of three.

(add ground capacity 6)

(table unit-size-in-terrain
  ;; (need other utypes also)
  (r-transp solid 3)
  (r-td solid 3)
  (r-tank solid 3)
  (g-misc solid 2)
  (g-td solid 2)
  (g-tank solid 2)
  ((block town) ground 6)  ; towns and blocks fill entire hex
  )

;;; Trucks, tanks, etc can carry one infantry or artillery unit.

(add u* capacity 1)
(add town capacity 6)
(add r-inf capacity 0)
(add g-inf capacity 0)
(add r-arty capacity 0)
(add g-arty capacity 0)

(table unit-size-as-occupant
  (u* u* 2)  ; can't be occupants usually
  (r-arty r-transp 1)
  (r-inf r-transp 1)
  (r-inf r-td 1)
  (r-inf r-tank 1)
  (g-arty g-transp 1)
  (g-inf g-transp 1)
  (g-inf g-misc 1)
  (g-inf g-td 1)
  (g-inf g-tank 1)
  (r-td town 3)
  (r-tank town 3)
  (g-td town 2)
  (g-tank town 2)
)

;;; Movement parameters.

(add u* acp-per-move 100)

(table mp-to-enter-terrain
  (u* t* 1)
  (u* slope 2)
  (u* road 0)
  )


(table mp-to-leave-terrain
  (u* t* 1)
  (r-td gully 3)
  (r-tank gully 3)
  (g-misc gully 3)
  (g-td gully 3)
  (g-tank gully 3)
  (u* road 0)
  )

(table mp-to-traverse
  (u* road 1)
  )

;;; Combat parameters.

(table acp-to-attack
  (u* u* 1) ; for now
  )

(table acp-to-defend
  (u* u* 1) ; for now
  )

;; In general, hp of 2 is healthy, 1 is crippled, 0 is dead or disappeared.
;; However, armored equipment can absorb more punishment.

(add u* hp-max 2)
(add town hp-max 30)
(add minefield hp-max 1)
(add r-td hp-max 4)
(add r-tank hp-max 4)
(add g-misc hp-max 4)
(add g-td hp-max 4)
(add g-tank hp-max 4)

(table hit-chance
  (u* u* 0)
  (r-arty u* 100)
  (r-inf g-inf 100)
  (r-td u* 100)
  (r-tank u* 100)
  (g-arty u* 100)
  (g-inf r-inf 100)
  (g-misc u* 100)
  (g-td u* 100)
  (g-tank u* 100)
  ;; Nobody can do anything to a burning wreck.
  (u* wreck 0)
  ;; Only engineers can clear minefields.
  (u* minefield 0)
  ((reng geng) minefield 20)
  )

(table damage
  (u* u* 0)
  (r-arty u* 1)
  (r-inf g-inf 1)
  (r-td u* 2)
  (su-152 u* 2)
  (r-tank u* 1)
  (g-arty u* 1)
  (g-inf r-inf 1)
  (g-misc u* 1)
  (wespe u* 3)
  (hummel u* 5)
  (g-td u* 1)
  (g-tank u* 1)
  (tiger1 u* 2)
  (u* wreck 0)  ; nobody can do anything to a burning wreck
  )

;;; Wrecked armored vehicles continue to be obstacles.

(add (rhalf ghalf) wrecked-type wreck)
(add r-td wrecked-type wreck)
(add r-tank wrecked-type wreck)
(add g-car wrecked-type wreck)
(add g-spa wrecked-type wreck)
(add g-td wrecked-type wreck)
(add g-tank wrecked-type wreck)

;;; Towns are easy to capture, that's why they should be protected.

(table capture-chance
  (u* town 100)
  )

;;; Visibility parameters.

(add u* vision-range 8)

(add wespe vision-range 32)

;;; The sides are pretty much fixed.

(side-library set
  ((name "Russia") (noun "Russian") (adjective "Russian") (class "russian")
   (color-scheme "red") (emblem-name "soviet-star") (names-locked true))
  ((name "Germany") (noun "German") (adjective "German") (class "german")
   (color-scheme "gray") (emblem-name "german-cross") (names-locked true))
  )

;;; Must always have exactly two sides.

(set sides-min 2)
(set sides-max 2)

;;; The following sets up a simple group of units for testing;
;;; usually "panzer" should be used in a well-defined scenario.

(add t* alt-percentile-min (  0  15  20  80  90   5   0   0   0))
(add t* alt-percentile-max ( 15  18  22  90 100 100   0   0   0))
(add t* wet-percentile-min (  0   0  50   0   0   0   0   0   0))
(add t* wet-percentile-max (100 100 100  50 100 100   0   0   0))

(set edge-terrain clear)

;;; (should make some sort of rational OB for matched sides)

(add town start-with 1)

(add (g75 r76at) start-with 2)

(add (g150 r122) start-with 1)

(add (t34 su-76 su-85 su-152) start-with (8 2 2 3))

(add (pz4 panther tiger1 wespe hummel) start-with (5 5 3 2 0))

(set country-radius-min 5)

(set country-separation-min 25)
(set country-separation-max 30)

;(add town density 1000)

(table road-chance
  (town town 25)
  )

(table road-into-chance
  (road t* 100)
  (road (water stream gully) 10)
  )

(scorekeeper (do last-side-wins))

(game-module (instructions "Move fast or die!"))

(game-module (design-notes (
  "world should have elevations, make shooting be LOS for armor, specify "
  "alts for artillery."
  )))

