/* Copyright (c) 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Functions common to all AIs. */

#include "conq.h"

char *aitypenames[] = {
  "brainless",
  "mplayer",
  NULL
};

init_ai(side)
Side *side;
{
    if (side_wants_ai(side)) {
	if (strcmp("mplayer", side->player->aitypename) == 0) {
	    mplayer_init(side);
	} else {
	}
	if (!side_has_ai(side)) {
	    init_warning("might not have made an AI (type %s) for %s",
			 side->player->aitypename, side_desig(side));
	}
    }
}

ai_init_turn(side)
Side *side;
{
    switch (side_ai_type(side)) {
      case nobrains:
	break;
      case mplayertype:
	mplayer_init_turn(side);
	break;
      default:
	case_panic("player AI type", side_ai_type(side));
	break;
    }
}

ai_decide_plan(side, unit)
Side *side;
Unit *unit;
{
    switch (side_ai_type(side)) {
      case nobrains:
	unit->plan->type = FOLLOW_ORDERS;
	unit->plan->tasks = NULL;
	break;
      case mplayertype:
	mplayer_decide_plan(side, unit);
	break;
      default:
	break;
    }
}

ai_react_to_unit_loss(side, unit)
Side *side;
Unit *unit;
{
    switch (side_ai_type(side)) {
      case mplayertype:
	mplayer_react_to_unit_loss(side, unit);
    }
}

ai_react_to_action_result(side, unit, rslt)
Side *side;
Unit *unit;
int rslt;
{
}

/* Goal and goalstate handling. */

struct a_goaldefn {
    char *name;
    char *argtypes;
} goaldefns[] = {

#undef  DEF_GOAL
#define DEF_GOAL(code,NAME,ARGTYPES) { NAME, ARGTYPES },

#include "goal.def"

    { NULL, NULL }
};


/* General handling of goals and goal states. */

Goal *
create_goal(type)
GoalType type;
{
    Goal *goal = (Goal *) xmalloc(sizeof(Goal));

    goal->type = type;
    return goal;
}

GoalState *
create_goalstate(goal, side2, state)
Goal *goal;
Side *side2;
int state;
{
    GoalState *goalstate = (GoalState *) xmalloc(sizeof(GoalState));

    goalstate->goal = goal;
    goalstate->side = side2;
    goalstate->state = state;
    goalstate->evaldate = curturn;
    goalstate->next = NULL;
    return goalstate;
}

hex_unknown(x, y)  { return (terrain_view(tmpside, x, y) == UNSEEN); }

#if 0
	    for_all_sides(side2) {
		if (enemy_side(side, side2) && !side2->lost) return -100;
	    }
#endif

/* Test a goal to see if it is true for side, as specified. */

goal_truth(side, side2, goal)
Side *side, *side2;
Goal *goal;
{
    int x, y;

    if (goal) {
	switch (goal->type) {
	  case LOST_GAME:
	    return (side2 ? (side_lost(side2) ? 100 : -100) : 0);
	  case POSITIONS_KNOWN:
	    /* what if no enemies present? then this is undefined? */
	    /* should goals have preconditions or prerequisites? */
	    return 0;
	  case WORLD_KNOWN:
	    if (side != side2) return 0;
	    tmpside = side;
	    for_all_interior_hexes(x, y) {
		if (hex_unknown(x, y)) return -100;
	    }
	    return 100;
	  case VICINITY_KNOWN:
	    tmpside = side;
	    if (search_around(goal->arg[0], goal->arg[1], goal->arg[2], hex_unknown,
			      &x, &y, 1)) {
		return -100;
	    } else {
		return 100;
	    }
	  default:
	    return 0;
	}
    } else {
	return 0;
    }
}

char goalbuf[BUFSIZE];

char *
goal_desig(goal)
Goal *goal;
{
    int i;
    char *argtypes;

    if (goal) {
	sprintf(goalbuf, "goal %s", goaldefns[goal->type].name);
	argtypes = goaldefns[goal->type].argtypes;
	for (i = 0; i < strlen(argtypes); ++i) {
	    switch (argtypes[i]) {
	      default:
		sprintf(goalbuf+strlen(goalbuf), " %d", goal->arg[i]);
	    }
	}
	return goalbuf;
    } else {
	return "null goal";
    }
}

char gsbuf[BUFSIZE];

char *
goalstate_desig(gs)
GoalState *gs;
{
    if (gs) {
	sprintf(gsbuf, "<s#%d %s state is %s (%d) turn %d>",
		side_number(gs->side),
		goal_desig(gs->goal),
		(gs->state > 0 ? "true" :
		 (gs->state == 0 ? "unknown" : "false")),
		gs->state, gs->evaldate);
	return gsbuf;
    } else {
	return "<null goalstate>";
    }
}
