/* Copyright (c) 1987, 1988, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Init of machine player data.  This is done at the beginning of a game. */

#include "conq.h"
#include "mplay.h"

#define set_uu_bhw(x,y,v)
#define set_uu_bcw(x,y,v)

int unit_count;

/* General collections of numbers used by all machine players. */

int can_move_in_dir[NUMDIRS];  /* directions it make sence to try to move in */

long bestworth = -10000, bestx, besty;

Unit *munit;                    /* Unit being decided about */

Side *mside;                    /* Side whose unit is being decided about */

/* General collections of numbers used by all machine players. */

int *fraction;        /* percentages of terrain types in area */
int *unit_hexes;      /* hexes of terrain in area unit can move on */
int *bw;              /* how much is a unit worth */
int *maxoccupant;     /* total capacity of a transport */
int *ave_build_time;  /* Average length of time to build a unit */
int *localworth;                /* for evaluation of nearby hexes */
#ifdef REGIONS
short **unit_region;  
     /* Regions a unit within which unit can move */
int **unit_region_size;
    /* How many hexes in each region.  Regions not necessarily contiguous,
       but decent estimate.  Allows lakes to be spotted. */
short *num_regions;   /* Number of regions for this unit type */
short **units_in_region; /* Number of units currently in region */
long munit_regions;            /* Regions current unit can reach */
#endif
Oparea *oparea_info;       /* info about areas, shared for all sides */
short opareas_wide;      /* number of areas in x direction */
short areas_high;      /* number of areas in y direction */
Oparea *top_opareas[NUMTOPOPAREAS];       /* Areas requesting the most units. */

int base_building = FALSE;     /* can some unit build a base */

int worths_known = FALSE;  /* does the period file have portions info */

/* Init used by all machine players.  Precompute useful information
   relating to unit types in general, and that usually gets referenced
   in inner loops. */

mplayer_init_shared()
{
    int u, u1, u2, t, r1, numbuilders;
    Side *side;
    int tmp;
    
    /* Need 3 scratch layers for routefinding. */
    allocate_area_scratch(3);

    localworth = malloc_area_layer(int);
    /* Compute approx percentage occurrence of each terrain type */
    /* in the area not yet seen */
    /* (not useful if fractal gen not used) */
    fraction = (int *) xmalloc(numttypes * sizeof(int));
    for_all_terrain_types(t) {
	fraction[t] = ((t_alt_max(t) - t_alt_min(t)) *
		       (t_wet_max(t) - t_wet_min(t))) / 100;
    }
    unit_hexes = (int *) xmalloc(numutypes * sizeof(int));
    bw = (int *) xmalloc(numutypes * sizeof(int));
    maxoccupant = (int *) xmalloc(numutypes * sizeof(int));
    ave_build_time = (int *) xmalloc(numutypes * sizeof(int));
#ifdef REGIONS
    unit_region = (short **) xmalloc(numutypes * sizeof(short *));
    unit_region_size = (int **) xmalloc(numutypes * sizeof(int *));
    num_regions = (short *) xmalloc(numutypes * sizeof(short));
    units_in_region = (short **) xmalloc(numutypes * sizeof(short *));
#endif
    /* Recognize unit types that are bases */
    for_all_unit_types(u1) {
	set_u_is_base(u1, FALSE);
	tmp = FALSE;
	for_all_material_types(r1) {
	    if (um_produce(u1, r1) > 0) {
		tmp = TRUE;
		continue;
	    }
	}
	if (tmp) {
	    for_all_unit_types(u2) {
		if ((u1 != u2) && could_carry(u1,u2)) {
		    set_u_is_base(u1, TRUE);
		    continue;
		}
	    }
	}
    }
    /* Note that is_base_builder is set to the type of base that can */
    /* be built.  That means that unit zero can not be a base which */
    /* can be built. */
    for_all_unit_types(u1) {
	if (u_defense_worth(u1) > 0)
	  worths_known = TRUE;
	set_u_is_transport(u1, FALSE);
	set_u_is_carrier(u1, FALSE);
	set_u_is_base_builder(u1, FALSE);
	set_u_can_make(u1, FALSE);
	set_u_can_capture(u1, FALSE);
	numbuilders = 0;
	ave_build_time[u1] = 0;
	for_all_unit_types(u2) {
	    if (u_is_base(u2) &&
		could_make(u1, u2) &&
		1 /* period_setproduct() */) {
		set_u_is_base_builder(u1, u2);
		base_building = TRUE;
	    }
	    if (u_speed(u1) > 0 && could_carry(u1,u2)) {
		set_u_is_transport(u1, TRUE);
	    }
	    if (could_make(u2, u1)) {
		numbuilders++;
/*		ave_build_time[u1] += uu_make(u2,u1);  */
		set_u_can_make(u2, TRUE);
	    }
	    if (could_capture(u1, u2)) {
		set_u_can_capture(u1, TRUE);
	    }
	}
	if (numbuilders > 0)
	  ave_build_time[u1] /= numbuilders;
    }
    /* a carrier is a unit that is a mobile base, but it can not
       move a passenger anywhere the passenger could not go itself. */
    for_all_unit_types(u1) {
	if (u_is_transport(u1)) {
	    set_u_is_carrier(u1, TRUE);
	    for_all_unit_types(u2) {
		if (could_carry(u1, u2)) {
		    for_all_terrain_types(t) {
			if (could_move(u1, t) && !could_move(u2, t))
			  set_u_is_carrier(u1, FALSE);
		    }
		}
	    }
	}
    }
    for_all_unit_types(u) {
	bw[u] = basic_worth(u);
    }
    /* repeat to get makers values right */
    /* still not right if makers can produce makers */

    for_all_unit_types(u)   
	bw[u] = basic_worth(u);
    for_all_unit_types(u) {
	maxoccupant[u] = 0;
	for_all_unit_types(u2) {
	    set_uu_bhw(u, u2, basic_hit_worth(u, u2));
	    set_uu_bcw(u, u2, basic_capture_worth(u, u2));
	    maxoccupant[u] += uu_capacity_x(u, u2);
	}
    }
#ifdef REGIONS
    determine_regions();
#endif
    /* tell us about how things rated */
    if (DebugM) display_assessment();
    init_areas();
}

/* A crude estimate of having one type of unit. */

int basic_worth(u)
int u;
{
    int worth = 0, u2, r, range;
  
    worth += u_hp(u) * 10;
    for_all_unit_types(u2) {
	if (could_make(u, u2))
	  worth += (bw[u2] * (50)) / 1 /* uu_make(u, u2) */;
	if (could_carry(u, u2))
	  worth += (1 + u_speed(u)) * uu_capacity_x(u, u2) *
	    (u_is_base(u) ? 10 : 1) * bw[u2] / 30;
    }
    range = 12345;
    for_all_material_types(r) {
	worth += um_produce(u, r) * (u_is_base(u) ? 4 : 1);
	if (um_tomove(u, r) > 0)
	  range = min(range, um_storage(u, r) / max(1, um_tomove(u, r)));
	if (um_consume(u, r) > 0) 
	  range =
	    min(range, u_speed(u) * um_storage(u, r) / max(1, um_consume(u, r)));
    }
    worth += u_speed(u) * u_hp(u);
    worth += (range == 12345 ? area.maxdim : range)
      * u_hp(u) / max(1, 10 - u_speed(u));
    for_all_unit_types(u2) {
	worth += (worth * uu_capture(u, u2)) / 150;
    }
    worth = isqrt(worth);
    DMprintf("unit type %s worth %d \n ", u_type_name(u), worth);
    return worth;
}

/* A crude estimate of the payoff of one unit type hitting on another type. */
/* This is just for general estimation, since actual worth may depend on */
/* damage already sustained, unit's goals, etc. */

basic_hit_worth(u, e)
int u, e;
{
    int worth, anti;

    worth = uu_hit(u,e) * min(u_hp(e), uu_damage(u, e));
    if (u_hp(e) > uu_damage(u, e)) {
	worth /= u_hp(e);
    }
    if (1 /* period_counterattack() */) {
	anti = uu_hit(e,u) * min(u_hp(u), uu_damage(u, e));
	if (u_hp(u) > uu_damage(u, e)) {
	    anti /= u_hp(u);
	}
    } else {
	anti = 0;
    }
    if (ave_build_time[u] > 0 && ave_build_time[e] > 0) {
	worth =  (worth * ave_build_time[e] - anti * ave_build_time[u]) /
	  (ave_build_time[u] + ave_build_time[e]);
    } else {
	worth -= anti;
    }
    return worth;
}

/* A crude estimate of the payoff of one unit type trying to capture. */

basic_capture_worth(u, e)
int u, e;
{
    int worth = 0;

    if (could_capture(u, e)) {
	worth = uu_capture(u, e) * bw[e];
    }
    return worth;
}

#ifdef REGIONS
/* Fill in a region */

void
fill_region(x, y, u, region, depth)
int x, y, u, region, depth;
{
    int d, nx, ny;

    if (depth > 1000) return;

    aset(unit_region[u], x, y, region);
    unit_hexes[u]++;
    for_all_directions(d) {
	nx = wrap(x + dirx[d]);  ny = y + diry[d];
	if (aref(unit_region[u], nx, ny) == -1 &&
	    could_move(u, terrain_at(nx, ny)))
	  fill_region(nx, ny, u, region, ++depth);
  }
}

/* Find continuous regions in which a unit can travel. */

determine_regions()
{
    int x, y, region, u, u1, i, t;
    long terr_profile[MAXUTYPES];
    int already_done[MAXUTYPES];
    int same_as[MAXUTYPES];
    int atv[MAXUTYPES];

    for_all_unit_types(u) {
	terr_profile[u] = 0;
	atv[u] = TRUE;
	for_all_terrain_types(t) {
	    if (could_move(u, t)) {
		terr_profile[u] |= 1<<t;
	    } else {
		atv[u] = FALSE;
	    }
	}
	for (u1 = 0; u1 < u; u1++) {
	    if (terr_profile[u] == terr_profile[u1])
	      break;
	}
	if (u1 < u) {
	    unit_region[u] = unit_region[u1];
	    if (DebugM)
	      DMprintf("Regions for %s's same as %s's\n",
		     u_type_name(u), u_type_name(u1));
	    already_done[u] = TRUE;
	    same_as[u] = u1;
	} else {
	    already_done[u] = FALSE;
	    unit_region[u] = malloc_area_layer(short);
	    for_all_interior_hexes(x,y)
	      if (between(1, y, area.height - 2)) {
		  aset(unit_region[u], x, y, (atv[u] ? 0 : -1));
	      } else {
		  aset(unit_region[u], x, y, -2);
	      }
	    /* top and bottom always off bounds */
	}
    }
    for_all_unit_types(u) {
	if (!already_done[u]) {
	    region = 0;
	    unit_hexes[u] = 0;
	    if (atv[u]) {
		region = 1;
		unit_hexes[u] = area.width * (area.height - 2);
	    } else {
		for_all_hexes(x,y)
		  if (aref(unit_region[u], x, y) == -1 &&
		      could_move(u, terrain_at(x, y))) {
		      DMprintf("Filling region %d for %s's\n",
					region, u_type_name(u));
		      fill_region(x, y, u, region++, 0);
		  }
	    }
	    num_regions[u] = region;
	    unit_region_size[u] = (int *) xmalloc(num_regions[u] * sizeof(int));
	    units_in_region[u] = (short *) xmalloc(num_regions[u] * sizeof(short));
	    for (i = 0; i < num_regions[u]; i++)
	      unit_region_size[u][i] = 0;
	    for_all_hexes(x, y)
	      if ((i = aref(unit_region[u], x, y)) >= 0)
		unit_region_size[u][i]++;
	    DMprintf("Found %d hexes for %s units\n",
			      unit_hexes[u], u_type_name(u));
	} else {
	    num_regions[u] = num_regions[same_as[u]];
	    unit_region_size[u] = unit_region_size[same_as[u]];
	    units_in_region[u] = (short *) xmalloc(num_regions[u] * sizeof(short));
	    unit_hexes[u] = unit_hexes[same_as[u]];
	}
    }
}
#endif

init_areas()
{
    int i, j, num;
    Side *side;
  
    opareas_wide = (area.width + CAREA_SIZE - 1) / CAREA_SIZE;
    areas_high = (area.height + CAREA_SIZE - 1) / CAREA_SIZE;
    oparea_info = (Oparea *) xmalloc(opareas_wide * areas_high * sizeof(Oparea));
    for_all_sides(side) {
	if (mplayer(side))
	mplayer(side)->areas =
	  (Control_area *) xmalloc(opareas_wide * areas_high * sizeof(Control_area));
    }
    for (i = 0; i < opareas_wide; i++) {
	for (j = 0; j < areas_high ; j++) {
	    num = j * opareas_wide + i;
	    oparea_info[num].number = num;
	    oparea_info[num].x = (CAREA_SIZE * i +
				min(CAREA_SIZE * (i+1), area.width)) / 2;
	    oparea_info[num].y = (CAREA_SIZE * j +
				min(CAREA_SIZE * (j+1), area.height)) / 2;
	    for_all_sides(side) {
		if (mplayer(side)) {
		    mplayer(side)->areas[num].enemy_strength = 0;
		    mplayer(side)->areas[num].enemy_seen_recently = 0;
		    mplayer(side)->areas[num].units_lost = 0;
		    mplayer(side)->areas[num].capturers_approaching = 0;
		    mplayer(side)->areas[num].capture_time = 10000;
		}
	    }
	}
    }
}

/* Print out everything that was calculated about the period. */
/* (Make this into a "designer's help node"? */

display_assessment()
{
    int t, u, u2;

    DMprintf("\n\nTerrain Percentages:\n");
    for_all_terrain_types(t) {
	DMprintf("%3d%% ", fraction[t]);
    }
    DMprintf("\nUnit Attributes:\n");
    for_all_unit_types(u) {
	DMprintf(" %s : base %d, transport %d, carrier %d, worth %d\n",
	       utype_name_n(u, 1), u_is_base(u),
	       u_is_transport(u), u_is_carrier(u), bw[u]);
    }
    DMprintf("\nUnit vs Unit Combat:\n");
    for_all_unit_types(u) {
	for_all_unit_types(u2) DMprintf("%5d", uu_bhw(u, u2));
	DMprintf("\n");
    }
    DMprintf("\nUnit vs Unit Capture:\n");
    for_all_unit_types(u) {
	for_all_unit_types(u2) DMprintf(" %4d", uu_bcw(u, u2));
	DMprintf("\n");
    }
    DMprintf("\n");
}

#define OPAREA_SIZE 10

mplayer_react_to_unit_loss(side, unit)
Side *side;
Unit *unit;
{
    int x = unit->x, y = unit->y;

    if (x < 0 || y < 0) {
	x = unit->prevx;  y = unit->prevy;
    }
    if (mplayer(side) && mplayer(side)->areas) {
	mplayer(side)->areas[oparea_index(x, y)].units_lost +=
	  1 /* severity */ * unit_strength(unit->type);
    }
}

#if 0 /* useful someday? */

/* Machine players ally with each other and make war on humans.
   Actually, the test is "host == NULL", in case a machine player
   is attached to a screen initially, we still want the others to fight
   with it.  In fact, all machines on displays will fight each other. */

machines_gang_up()
{
    Side *side, *side2;

    for_all_sides(side) {
	for_all_sides(side2) {
	    if (neutral_side(side, side2)) {
		if (side->displayname != NULL || side2->displayname != NULL) {
		    declare_war(side, side2);
		} else {
		    declare_alliance(side, side2);
		}
	    }
	}
    }
}
#endif

set_u_is_base(u, n) int u, n; {  utypes[u].is_base = n;  }
set_u_is_transport(u, n) int u, n; {  utypes[u].is_transport = n;  }
set_u_is_carrier(u, n) int u, n; {  utypes[u].is_carrier = n;  }
set_u_is_base_builder(u, n) int u, n; {  utypes[u].is_base_builder = n;  }
set_u_can_make(u, n) int u, n; {  utypes[u].can_make = n;  }
set_u_can_capture(u, n) int u, n; {  utypes[u].can_capture = n;  }

