/* Copyright (c) 1987-1993  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Game module writing. */

#include "conq.h"

/* (should make the fp a global, would shrink code significantly) */

#define start_form(fp,hd) (fprintf((fp), "(%s", (hd)))
#define add_to_form(fp,x) (fprintf((fp), " %s", (x)))
#define end_form(fp) (fprintf((fp), ")"))
#define newline_form(fp) (fprintf((fp), "\n"))
#define space_form(fp) (fprintf((fp), " "))

extern int gamestatesafe;

int doreshape = FALSE;

Module *reshaper = NULL;

/* These two routines make sure that any symbols and strings can
   be read in again. */

char *escapedthing = NULL;

char *
escaped_symbol(str)
char *str;
{
    char *tmp = str;

    while (*tmp != '\0') {
	if (((char *) strchr(" ()#\";|", *tmp)) != NULL) {
	    sprintf(escapedthing, "|%s|", str);
	    return escapedthing;
	}
	++tmp;
    }
    return str;
}

/* Note that this works correctly on NULL strings, turning them into
   strings of length 0. */

char *
escaped_string(str)
char *str;
{
    char *tmp = str, *rslt = escapedthing;

    *rslt++ = '"';
    if (str != NULL) {
	while (*tmp != 0) {
	    if (*tmp == '"') *rslt++ = '\\';
	    *rslt++ = *tmp++;
	}
    }
    *rslt++ = '"';
    *rslt = '\0';
    return escapedthing;
}

write_boolean_property(fp, name, value/*, dflt, usedflt, addnewline*/)
FILE *fp;
char *name;
int value;
{
    fprintf(fp, " (%s %s)", name, (value ? "true" : "false"));
}

write_number_property(fp, name, value/*, dflt, usedflt, addnewline*/)
FILE *fp;
char *name;
int value;
{
    fprintf(fp, " (%s %d)", name, value);
}

/* Handle the writing of a single string-valued property. */

write_string_property(fp, name, value/*, dflt, usedflt, addnewline*/)
FILE *fp;
char *name, *value;
{
    fprintf(fp, "  (%s %s)", name, escaped_string(value));
}

write_lisp_property(fp, name, value/*, dflt, usedflt, addnewline*/)
FILE *fp;
char *name;
Obj *value;
{
    fprintf(fp, "  (%s xxx)", name/*, escaped_string(value)*/);
}

char *shortestbuf = NULL;

char *
shortest_unique_name(u)
int u;
{
    char *typename = u_type_name(u);

	if (shortestbuf == NULL) shortestbuf = (char *) xmalloc(BUFSIZE * sizeof(char));
    if (strlen(u_internal_name(u)) < strlen(typename)) {
	typename = u_internal_name(u);
    }
    sprintf(shortestbuf, "%s", escaped_symbol(typename));
    return shortestbuf;
}

/* A saved game should include everything necessary to recreate a game
   exactly.  It is important that this routine not attempt to use graphics,
   since it may be called when graphics code fails. */

write_entire_game_state(fname)
char *fname;
{
    Module *module = create_game_module(fname);
    int rslt;

    module->filename = fname;
    module->compresslayers = TRUE;
    module->defall = TRUE;
    rslt = write_game_module(module);
    if (rslt) gamestatesafe = TRUE;
    return rslt;
}

/* Given a game module telling what is in the module, write out a file
   containing the requested content.  Return true if everything went OK. */

write_game_module(module)
Module *module;
{
    int i;
    FILE *fp;
    Obj *rest;

    if (escapedthing == NULL) escapedthing = (char *) xmalloc(BUFSIZE * sizeof(char));
    if ((fp = fopen(module->filename, "w")) != NULL) {
	/* Write the definition of this game module. */
	start_form(fp, "game-module");  space_form(fp);
	add_to_form(fp, escaped_string(module->name));  newline_form(fp);
	/* print blurb etc eventually */
	if (1 /* module->defall */) {
	    write_string_property(fp, "program-version", version);
	}
	end_form(fp);  newline_form(fp);  newline_form(fp);
	if (module->basemodulename) { /* should go in game module? */
	    start_form(fp, "include");
	    add_to_form(fp, escaped_string(module->basemodulename));
	    end_form(fp);  newline_form(fp);
	}
	if (module->defall || module->deftypes)
	  write_types(fp, module);
	if (module->defall || module->deftables)
	  write_tables(fp, module);
	if (module->defall || module->defglobals)
	  write_globals(fp, module);
	if (module->defall || module->defscoring)
	  write_scorekeepers(fp, module);
	doreshape = reshape_the_output(module);
	reshaper = module;
	if (module->defall || module->defworld)
	  write_world(fp, module);
	if (module->defall || module->defareas)
	  write_areas(fp, module);
	if (module->defall || module->defsides)
	  write_sides(fp, module);
	if (module->defall || module->defplayers)
	  write_players(fp, module);
	if (module->defall || module->defunits)
	  write_units(fp, module, module);
	if (module->defall || module->defhistory)
	  write_history(fp, module);
	/* Write the game notes here (seems reasonable, no deeper reason). */
	if (module->instructions != lispnil) {
	    fprintf(fp, "(game-module (instructions (\n");
	    for (rest = module->instructions; rest != lispnil; rest = cdr(rest)) {
		fprintlisp(fp, car(rest));
	    }
	    fprintf(fp, "  )))\n");
	}
	if (module->notes != lispnil) {
	    fprintf(fp, "(game-module (notes (\n");
	    for (rest = module->notes; rest != lispnil; rest = cdr(rest)) {
		fprintlisp(fp, car(rest));
	    }
	    fprintf(fp, "  )))\n");
	}
	if (module->designnotes != lispnil) {
	    fprintf(fp, "(game-module (design-notes (\n");
	    for (rest = module->designnotes; rest != lispnil; rest = cdr(rest)) {
		fprintlisp(fp, car(rest));
	    }
	    fprintf(fp, "  )))\n");
	}
	fclose(fp);
	return TRUE;
    } else {
	return FALSE;
    }
}

/* Write definitions of all the types. */

write_types(fp, module)
FILE *fp;
Module *module;
{
    int complete = FALSE;
    int u, m, t, i, ival;
    char *sval;
    Obj *obj;

    /* (or write out all the default values first for doc, then
       only write changed values) */

    for_all_unit_types(u) {
	fprintf(fp, "(unit-type %s\n", shortest_unique_name(u));
	for (i = 0; utypedefns[i].name != NULL; ++i) {
	    if (utypedefns[i].intgetter) {
		ival = (*(utypedefns[i].intgetter))(u);
		if (ival != utypedefns[i].dflt) {
		    fprintf(fp, "  (%s %d)\n", utypedefns[i].name, ival);
		}
	    } else if (utypedefns[i].strgetter) {
		sval = (*(utypedefns[i].strgetter))(u);
		if (string_not_default(sval, utypedefns[i].dfltstr)) {
		    write_string_property(fp, utypedefns[i].name, sval);
		    fprintf(fp, "\n");
		}
	    } else {
		obj = (*(utypedefns[i].objgetter))(u);
		if (obj != lispnil) {
		    fprintf(fp, "  (%s ", utypedefns[i].name);
		    fprintlisp(fp, obj);
		    fprintf(fp, ")\n");
		}
	    }
	}
	fprintf(fp, ")\n");
    }
    for_all_material_types(m) {
	fprintf(fp, "(material-type %s\n", escaped_symbol(m_type_name(m)));
	for (i = 0; mtypedefns[i].name != NULL; ++i) {
	    if (mtypedefns[i].intgetter) {
		ival = (*(mtypedefns[i].intgetter))(m);
		if (ival != mtypedefns[i].dflt) {
		    fprintf(fp, "  (%s %d)\n", mtypedefns[i].name, ival);
		}
	    } else if (mtypedefns[i].strgetter) {
		sval = (*(mtypedefns[i].strgetter))(m);
		if (string_not_default(sval, mtypedefns[i].dfltstr)) {
		    write_string_property(fp, mtypedefns[i].name, sval);
		    fprintf(fp, "\n");
		}
	    } else {
		obj = (*(mtypedefns[i].objgetter))(m);
		if (obj != lispnil) {
		    fprintf(fp, "  (%s ", mtypedefns[i].name);
		    fprintlisp(fp, obj);
		    fprintf(fp, ")\n");
		}
	    }
	}
	fprintf(fp, ")\n");
    }
    for_all_terrain_types(t) {
	fprintf(fp, "(terrain-type %s\n", escaped_symbol(t_type_name(t)));
	for (i = 0; ttypedefns[i].name != NULL; ++i) {
	    if (ttypedefns[i].intgetter) {
		ival = (*(ttypedefns[i].intgetter))(t);
		if (ival != ttypedefns[i].dflt) {
		    fprintf(fp, "  (%s %d)\n", ttypedefns[i].name, ival);
		}
	    } else if (ttypedefns[i].strgetter) {
		sval = (*(ttypedefns[i].strgetter))(t);
		if (string_not_default(sval, ttypedefns[i].dfltstr)) {
		    write_string_property(fp, ttypedefns[i].name, sval);
		    fprintf(fp, "\n");
		}
	    } else {
		obj = (*(ttypedefns[i].objgetter))(t);
		if (obj != lispnil) {
		    fprintf(fp, "  (%s ", ttypedefns[i].name);
		    fprintlisp(fp, obj);
		    fprintf(fp, ")\n");
		}
	    }
	}
	fprintf(fp, ")\n");
    }
    fprintf(fp, "\n");
}

string_not_default(str, dflt)
char *str, *dflt;
{
    if (dflt == NULL || strlen(dflt) == 0) {
	if (str == NULL || strlen(str) == 0) {
	    return FALSE;
	} else {
	    return TRUE;
	}
    } else {
	if (str == NULL || strlen(str) == 0) {
	} else {
	    return (strcmp(str, dflt) != 0);
	}
    }
}

/* Write definitions of all the tables. */

write_tables(fp, module)
FILE *fp;
Module *module;
{
    int tbl;

	fprintf(fp, "\n");
    for (tbl = 0; tabledefns[tbl].name != 0; ++tbl) {
	if (*(tabledefns[tbl].table) != NULL) {
	    write_table(fp, tabledefns[tbl].name,
			tabledefns[tbl].getter, tabledefns[tbl].dflt,
			tabledefns[tbl].index1, tabledefns[tbl].index2);
	}
    }
}

/* Write out a single table.  Only write it if it contains non-default
   values, and try to find runs of constant value. */

/* (should merge multiple rows also) */

#define star_from_typ(typ)  \
  ((typ) == UTYP ? "u*" : ((typ) == MTYP ? "m*" :"t*"))

#define name_from_typ(typ, i)  \
  ((typ) == UTYP ? u_type_name(i) : ((typ) == MTYP ? m_type_name(i) : t_type_name(i)))

write_table(fp, name, getter, dflt, typ1, typ2)
FILE *fp;
char *name;
int (*getter)(), dflt, typ1, typ2;
{
    int i, j, rowvalue, colvalue, constrow, constcol;
    int dim1 = numtypes_from_index_type(typ1);
    int dim2 = numtypes_from_index_type(typ2);

    fprintf(fp, "(table %s #| %d |#", name, dflt);
  if (dim1 <= dim2) {
    for (i = 0; i < dim1; ++i) {
	constrow = TRUE;
	rowvalue = (*getter)(i, 0);
	for (j = 0; j < dim2; ++j) {
	    if ((*getter)(i, j) != rowvalue) {
		constrow = FALSE;
		break;
	    }
	}
	if (!constrow || rowvalue != dflt) {
	    fprintf(fp, "\n  (%s %s",
		    escaped_symbol(name_from_typ(typ1, i)),
		    star_from_typ(typ2));
	    if (constrow) {
		fprintf(fp, " %d", rowvalue);
	    } else {
		fprintf(fp, " (");
		for (j = 0; j < dim2; ++j) {
		    fprintf(fp, " %d", (*getter)(i, j));
		}
		fprintf(fp, ")");
	    }
	    fprintf(fp, ")");
	}
    }
  } else {
    for (i = 0; i < dim2; ++i) {
	constcol = TRUE;
	colvalue = (*getter)(0, i);
	for (j = 0; j < dim1; ++j) {
	    if ((*getter)(j, i) != colvalue) {
		constcol = FALSE;
		break;
	    }
	}
	if (!constcol || colvalue != dflt) {
	    fprintf(fp, "\n  (%s %s",
		    star_from_typ(typ1),
		    escaped_symbol(name_from_typ(typ2, i)));
	    if (constcol) {
		fprintf(fp, " %d", colvalue);
	    } else {
		fprintf(fp, " (");
		for (j = 0; j < dim1; ++j) {
		    fprintf(fp, " %d", (*getter)(j, i));
		}
		fprintf(fp, ")");
	    }
	    fprintf(fp, ")");
	}
    }
  }
    fprintf(fp, ")\n");
}

/* Write info about the whole world. */

write_world(fp, module)
FILE *fp;
Module *module;
{
    fprintf(fp, "\n(world %d",
	    (doreshape ? reshaper->finalcircumference : world.circumference));
    if (world.daylength != 1)
      write_number_property(fp, "day-length", world.daylength);
    if (world.yearlength != 1)
      write_number_property(fp, "year-length", world.yearlength);
    fprintf(fp, ")\n");
}

/* Write info about the area in the world.  This code uses run-length encoding
   to reduce the size of each written layer as much as possible.  Note
   also that each layer is written as a separate form, so that the Lisp
   reader doesn't have to read really large forms back in. */

write_areas(fp, module)
FILE *fp;
Module *module;
{
	int all = module->defall, compress = module->compresslayers;

    /* Write the basic dimensions. */
	fprintf(fp, "\n(area %d %d",
		(doreshape ? reshaper->finalwidth : area.width),
		(doreshape ? reshaper->finalheight : area.height));
    /* Write all the scalar properties. */
    if (area.latitude != 0)
      write_number_property(fp, "latitude", area.latitude);
    if (area.longitude != 0)
      write_number_property(fp, "longitude", area.longitude);
    if (area.cellwidth != 0)
      write_number_property(fp, "cell-width", area.cellwidth);
    fprintf(fp, ")\n");
    /* Write the area's layers, each as a separate form. */
    if (all || module->defareaterrain) write_area_terrain(fp, compress);
    if (all || module->defareaterrain) write_area_aux_terrain(fp, compress);
    if (all || module->defareamisc) write_area_features(fp, compress);
    if (all || module->defareamisc) write_area_elevations(fp, compress);
    if (all || module->defareamisc) write_area_people_sides(fp, compress);
    if (all || module->defareamaterial) write_area_materials(fp, compress);
    if (all || module->defareaweather) write_area_temperatures(fp, compress);
    if (all || module->defareaweather) write_area_clouds(fp, compress);
    if (all || module->defareaweather) write_area_winds(fp, compress);
}

fn_terrain_at(x, y)
int x, y;
{
    return terrain_at(x, y);
}

write_area_terrain(fp, compress)
FILE *fp;
int compress;
{
	int t;

    fprintf(fp, "(area (terrain\n  (by-name");
    for_all_terrain_types(t) {
    	fprintf(fp, " (%d %s)", t, escaped_symbol(t_type_name(t)));
    	if (t > 0 && t % 8 == 0) fprintf(fp, "\n          ");
    }
    fprintf(fp, ")\n");
    write_rle(fp, fn_terrain_at, 0, numttypes-1, NULL, compress);
    fprintf(fp, "))\n");
}

fn_aux_terrain_at(x, y)
int x, y;
{
    return aux_terrain_at(x, y, tmpttype);
}

write_area_aux_terrain(fp, compress)
FILE *fp;
int compress;
{
    int t;

    for_all_terrain_types(t) {
	if (aux_terrain_defined(t)) {
	    fprintf(fp, "(area (aux-terrain %s\n",
		    escaped_symbol(t_type_name(t)));
	    tmpttype = t;
	    write_rle(fp, fn_aux_terrain_at, 0, 127, NULL, compress);
	    fprintf(fp, "))\n");
	}
    }
}

fn_feature_at(x, y)
int x, y;
{
    return raw_feature_at(x, y);
}

write_area_features(fp, compress)
FILE *fp;
int compress;
{
    GeoFeature *feature;
    extern GeoFeature *featurelist;

    if (featurelist == NULL || !features_defined()) return;
    fprintf(fp, "(area (features (\n");
    /* Dump out the list of features first. */
    for (feature = featurelist; feature != NULL; feature = feature->next) {
	fprintf(fp, "   (%d %s",
		feature->id,
		escaped_string(feature->typename));
	fprintf(fp, " %s)\n",
		escaped_string(feature->name));
    }
    fprintf(fp, "  )\n");
    /* Now record which features go with which hexes. */
    write_rle(fp, fn_feature_at, 0, -1, NULL, compress);
    fprintf(fp, "))\n");
}

fn_elevation_at(x, y)
int x, y;
{
    return elev_at(x, y);
}

write_area_elevations(fp, compress)
FILE *fp;
int compress;
{
  if (elevations_defined()) {
    fprintf(fp, "(area (elevations\n");
    write_rle(fp, fn_elevation_at, minelev, maxelev, NULL, compress);
    fprintf(fp, "))\n");
  }
}

fn_people_side_at(x, y)
int x, y;
{
    return people_side_at(x, y);
}

write_area_people_sides(fp, compress)
FILE *fp;
int compress;
{
  if (people_sides_defined()) {
    fprintf(fp, "(area (people-sides\n");
    write_rle(fp, fn_people_side_at, 0, MAXSIDES+1, NULL, compress);
    fprintf(fp, "))\n");
  }
}

fn_material_at(x, y)
int x, y;
{
    return material_at(x, y, tmpmtype);
}

write_area_materials(fp, compress)
FILE *fp;
int compress;
{
    int m;

  if (any_cell_materials_defined()) {
    for_all_material_types(m) {
	if (cell_material_defined(m)) {
	    fprintf(fp, "(area (material %s\n",
		    escaped_symbol(m_type_name(m)));
	    tmpmtype = m;
	    write_rle(fp, fn_material_at, 0, 127, NULL, compress);
	    fprintf(fp, "))\n");
	}
    }
  }
}

fn_temperature_at(x, y)
int x, y;
{
    return temperature_at(x, y);
}

write_area_temperatures(fp, compress)
FILE *fp;
int compress;
{
  if (temperatures_defined()) {
    fprintf(fp, "(area (temperatures\n");
    write_rle(fp, fn_temperature_at, 0, 9999, NULL, compress);
    fprintf(fp, "))\n");
  }
}

write_area_clouds(fp, compress)
FILE *fp;
int compress;
{
  if (0 /* clouds_defined() */) {
    fprintf(fp, "(area (clouds\n");
    write_rle(fp, NULL, 0, 9999, NULL, compress);
    fprintf(fp, "))\n");
  }
}

write_area_winds(fp, compress)
FILE *fp;
int compress;
{
  if (0 /* winds_defined() */) {
    fprintf(fp, "(area (winds\n");
    write_rle(fp, NULL, 0, 9999, NULL, compress);
    fprintf(fp, "))\n");
  }
}

/* Write the globals.  The "complete" flag forces all values out, even
   if they match the compiled-in defaults. */

write_globals(fp, module)
FILE *fp;
Module *module;
{
    int complete = FALSE;

	fprintf(fp, "\n");

#undef  DEF_VAR_I
#define DEF_VAR_I(STR,FNAME,setfname,doc,var,lo,DFLT,hi)  \
    if (complete || FNAME() != DFLT)  \
      fprintf(fp, "(set %s %d)\n", STR, FNAME());
#undef  DEF_VAR_S
#define DEF_VAR_S(STR,FNAME,setfname,doc,var,DFLT)  \
    if (complete || string_not_default(FNAME(), DFLT))  \
      fprintf(fp, "(set %s %s)\n", STR, escaped_string(FNAME()));
#undef  DEF_VAR_L
#define DEF_VAR_L(STR,FNAME,setfname,doc,var)  \
    if (complete || FNAME() != lispnil)  {\
        fprintf(fp, "(set %s ", STR);  \
        fprintlisp(fp, FNAME());  \
        fprintf(fp, ")\n");  \
    }

#include "gvar.def"

}

write_scorekeepers(fp, module)
FILE *fp;
Module *module;
{
    int i, u, r;
    Scorekeeper *sk;

    for_all_scorekeepers(sk) {
	fprintf(fp, "(scorekeeper");
	fprintf(fp, ")\n");
    }
}

/* Write declarations of all the sides. */

write_sides(fp, module)
FILE *fp;
Module *module;
{
    Side *side;

    fprintf(fp, "\n; %d sides\n", numsides);
    Dprintf("Will try to write %d sides ...\n", numsides);
    for_all_sides(side) {
	fprintf(fp, "(side %d", side->id);
	write_side_properties(side, fp);
	fprintf(fp, "  )\n");
	if (module->defall || module->defsideviews) write_side_view(side, fp);
	Dprintf("Wrote side %s,\n", side_desig(side));
    }
    Dprintf("... Done writing sides.\n");
}

/* Write random properties of a side. */

write_side_properties(side, fp)
Side *side;
FILE *fp;
{
    int u, s, i;
    extern int numscores;

    write_string_property(fp, "name", side->name);
    write_string_property(fp, "long-name", side->longname);
    write_string_property(fp, "short-name", side->shortname);
    write_string_property(fp, "noun", side->noun);
    write_string_property(fp, "plural-noun", side->pluralnoun);
    write_string_property(fp, "adjective", side->adjective);
    write_string_property(fp, "noun", side->noun);
    write_string_property(fp, "plural-noun", side->pluralnoun);
    write_string_property(fp, "adjective", side->adjective);
    write_string_property(fp, "color", side->colorscheme);
    write_string_property(fp, "emblem-name", side->emblemname);
    write_string_property(fp, "class", side->sideclass);
    write_boolean_property(fp, "active", side->ingame);
    write_number_property(fp, "status", side->status);
    write_number_property(fp, "turn-time-used", side->turntimeused);
    write_number_property(fp, "total-time-used", side->totaltimeused);
    write_number_property(fp, "finished-turn", side->finishedturn);
    write_number_property(fp, "designer", side->designer);
    write_number_property(fp, "controlled-by", side_number(side->controlledby));
    fprintf(fp, "\n");
    fprintf(fp, "  (trusts");
    for (s = 0; s < numsides+1; ++s) fprintf(fp, " %d", side->trusts[s]);
    fprintf(fp, ")\n");
    fprintf(fp, "  (trades");
    for (s = 0; s < numsides+1; ++s) fprintf(fp, " %d", side->trades[s]);
    fprintf(fp, ")\n");
    fprintf(fp, "  (next-numbers");
    for_all_unit_types(u) fprintf(fp, " %d", side->counts[u]);
    fprintf(fp, ")\n");
    fprintf(fp, "  (tech");
    for_all_unit_types(u) fprintf(fp, " %d", side->tech[u]);
    fprintf(fp, ")\n");
    fprintf(fp, "  (init-tech");
    for_all_unit_types(u) fprintf(fp, " %d", side->inittech[u]);
    fprintf(fp, ")\n");
    if (side->scores) {
	fprintf(fp, "  (scores");
	for (i = 0; i < numscores; ++i) {
	    fprintf(fp, " %d", side->scores[i]);
	}
	fprintf(fp, ")\n");
    }
}

/* Write about what has been seen in the area. (should be more compact) */
/* (should have option to spec symbolic dict of sides and units) */

fn_terrain_view(x, y)
int x, y;
{
    return terrain_view(tmpside, x, y);
}

fn_unit_view(x, y)
int x, y;
{
    return unit_view(tmpside, x, y);
}

fn_unit_view_date(x, y)
int x, y;
{
    return 0 /* side_view_timestamp(tmpside, x, y) */ ;
}

write_side_view(side, fp, compress)
Side *side;
FILE *fp;
int compress;
{
    if (g_see_all()) return;
    tmpside = side;
    /* (should skip if all terrain seen) */
    fprintf(fp, "(side %d (terrain-view\n", side_number(side));
    write_rle(fp, fn_terrain_view, 0, 10000, NULL, compress);
    fprintf(fp, "  ))\n");
    fprintf(fp, "(side %d (unit-view\n", side_number(side));
    write_rle(fp, fn_unit_view, 0, 10000, NULL, compress);
    fprintf(fp, "  ))\n");
    fprintf(fp, "(side %d (unit-view-date\n", side_number(side));
    write_rle(fp, fn_unit_view_date, 0, 10000, NULL, compress);
    fprintf(fp, "  ))\n");
}

write_players(fp, module)
FILE *fp;
Module *module;
{
    Side *side;
    Player *player;

    fprintf(fp, "; %d players\n", numplayers);
    Dprintf("Will try to write %d players ...\n", numplayers);
    for_all_sides(side) {
	if ((player = side->player) != NULL) {
	    write_player(fp, player);
	    Dprintf("Wrote player %s,\n", player_desig(player));
	}
    }
    Dprintf("... Done writing players.\n");
}

write_player(fp, player)
FILE *fp;
Player *player;
{
    if (player == NULL) return;
    fprintf(fp, "(player %d\n", 0);
    write_string_property(fp, "name", player->name);
    write_string_property(fp, "config-name", player->configname);
    write_string_property(fp, "display-name", player->displayname);
    write_string_property(fp, "ai-type-name", player->aitypename);
    fprintf(fp, "  )\n");
}

/* Should write out "unit groups" with dict prepended, then can use with
   multiple games */

/* Write the unit section of a game module. */

write_units(fp, module)
FILE *fp;
Module *module;
{
    int x, y;
    Unit *unit;
    Side *loopside;

    /* Make sure no dead units get saved. */
    flush_dead_units();
    fprintf(fp, "; %d units\n", numunits);
    /* Need to write out the defaults being assumed subsequently. */
    /* maybe use those in postprocessing. */
    fprintf(fp, "(unit-defaults)\n");
    Dprintf("Writing %d units ...\n", numunits);
    for_all_units(loopside, unit) {
	if (1 /* test unit for worthiness to be written */) {
	    if (doreshape) {
		reshaped_point(unit->x, unit->y, &x, &y);
	    } else {
		x = unit->x;  y = unit->y;
	    }
	    fprintf(fp, "(%s %d %d %d",
		    shortest_unique_name(unit->type),
		    x, y, side_number(unit->side));
	    if (unit->z != 0) {
		fprintf(fp, " (z %d)", unit->z);
	    }
	    if (unit->name) {
		write_string_property(fp, "n", unit->name);
	    }
	    /* Only need if there are occupants (or other xrefs?) */
	    if (unit->occupant) {
		write_number_property(fp, "#", unit->id);
	    }
	    /* Need this to get back into the right transport. */
	    if (unit->transport) {
		write_number_property(fp, "in", unit->transport->id);
	    }
	    /* Write optional info about the units. */
	    if (module->defall || module->defunitprops)
	      write_unit_properties(unit, fp);
	    if (module->defall || module->defunitacts)
	      write_unit_act(unit, fp);
	    if (module->defall || module->defunitplans)
	      write_unit_plan(unit, fp);
	    /* close the unit out */
	    fprintf(fp, ")\n");
	    Dprintf("Wrote %s\n", unit_desig(unit));
	}
    }
    Dprintf("... Done writing units.\n");
}

/* Write random properties, but only if they have non-default values. */

write_unit_properties(unit, fp)
Unit *unit;
FILE *fp;
{
    int m, allfull = TRUE;

    if (unit->number != 0) {
	write_number_property(fp, "nm", unit->number);
    }
    if (unit->hp < u_hp(unit->type)) {
	write_number_property(fp, "hp", unit->hp);
    }
    if (unit->cp < u_cp(unit->type)) {
	write_number_property(fp, "cp", unit->cp);
    }
    if (nummtypes > 0) {
	for_all_material_types(m) {
	    if (unit->supply[m] < um_storage(unit->type, m)) {
		allfull = FALSE;
		break;
	    }
	}
	if (!allfull) {
	    fprintf(fp, " (m");
	    if (1 /* write all of them */) {
		for_all_material_types(m) {
		    fprintf(fp, " %d", unit->supply[m]);
		}
	    } else {
	        /* (should write a named list) */
	    }
	    fprintf(fp, ")");
	}
    }
}

/* Write out the unit's current actor state. */

write_unit_act(unit, fp)
Unit *unit;
FILE *fp;
{
    ActorState *act = unit->act;

    if (act) {
	fprintf(fp, " (act");
	write_number_property(fp, "acp", act->acp);
	write_number_property(fp, "acp0", act->initacp);
	/* (should dump action if not executed yet) */
	fprintf(fp, ")");
    }
}

/* Write out the unit's current plan. */

write_unit_plan(unit, fp)
Unit *unit;
FILE *fp;
{
    int i;
    Task *task;
    Plan *plan = unit->plan;
    extern char *plantypenames[];
    extern char *tasktypenames[];

    if (plan) {
	fprintf(fp, " (plan %s", plantypenames[plan->type]);
	if (plan->asleep)
	  write_boolean_property(fp, "asleep", plan->asleep);
	if (plan->reserve)
	  write_boolean_property(fp, "reserve", plan->reserve);
	if (plan->waitingfortasks)
	  write_boolean_property(fp, "wait", plan->waitingfortasks);
	if (plan->maingoal) {
	    fprintf(fp, " (goal");
	    fprintf(fp, ")");
	}
	if (plan->tasks) {
	    fprintf(fp, " (tasks");
	    for (task = plan->tasks; task != NULL; task = task->next) {
		fprintf(fp, " (%s %d %d", tasktypenames[task->type],
			task->execnum, task->retrynum);
		for (i = 0; i < 4; ++i) {
		    fprintf(fp, " %d", task->args[i]);
		}
		fprintf(fp, ")");
	    }
	    fprintf(fp, ")");
	}
	fprintf(fp, ")");
    }
}

/* This is a generalized routine to do run-length-encoding of area layers.
   It uses hook fns to acquire data at a point and an optional translator to
   do any last-minute fixing.  It can use either a char or numeric encoding,
   depending on the expected range of values. */

/* The default translator. */

fn_identity(x) int x; { return x; }

write_rle(fp, datafn, lo, hi, translator, compress)
FILE *fp;
int (*datafn)(), lo, hi, (*translator)(), compress;
{
    int width, height, x, y, x0, y0, run, runval, val, trval, i, iter;

    if (translator == NULL) translator = fn_identity;
    width = area.width;  height = area.height;
    if (doreshape) {
	width = reshaper->finalwidth;  height = reshaper->finalheight;
    }
    for (y = height-1; y >= 0; --y) {
	fprintf(fp, "  \"");
	run = 0;
	x0 = 0;  y0 = y;
	if (doreshape) {
	    original_point(0, y, &x0, &y0);
	}
	val = runval = (*datafn)(x0, y0);
	for (x = 0; x < width; ++x) {
	    x0 = x;  y0 = y;
	    if (doreshape) {
		original_point(x, y, &x0, &y0);
	    }
	    val = (*datafn)(x0, y0);
	    if (val == runval && compress) {
		run++;
	    } else {
		trval = (*translator)(runval);
		write_run(fp, run, trval);
		runval = val;
		run = 1;
	    }
	}
	/* Finish off the row. */
	trval = (*translator)(val);
	write_run(fp, run, trval);
	fprintf(fp, "\"\n");
    }
}

/* Write a single run, using the most compact encoding possible. */
/* 0 - 29 is 'a' - '~', 30 - 63 is ':' - '[' */ 

write_run(fp, run, val)
FILE *fp;
int run, val;
{
    if (run > 1) {
	fprintf(fp, "%d", run);
	if (val > 63) fprintf(fp, "*");
    }
    if (between(0, val, 29)) {
	fprintf(fp, "%c", val + 'a');
    } else if (between(30, val, 63)) {
	fprintf(fp, "%c", val - 30 + ':');
    } else {
	fprintf(fp, "%d,", val);
    }
}

write_history(fp, selector)
FILE *fp;
char *selector;
{
    HistEvent *hevt;

    write_historical_event(fp, history);
    for (hevt = history->next; hevt != history; hevt = hevt->next) {
	write_historical_event(fp, hevt);
    }
}

write_historical_event(fp, hevt)
FILE *fp;
HistEvent *hevt;
{
    fprintf(fp, "(hist %d %s", hevt->startdate, hevtdefns[hevt->type].name);
    switch (hevt->type) {
      case H_LOG_STARTED:
	break;
      case H_GAME_STARTED:
	break;
      case H_SIDE_JOINED:
      case H_SIDE_LOST:
      case H_SIDE_WON:
	if (hevt->data) {
	    fprintf(fp, " %d", ((Side *) hevt->data)->id);
	} else {
	    fprintf(fp, "0 #| ??? |#");
	}
	break;
      default:
	break;
    }
    fprintf(fp, ")\n");
}

/* Statistics are a compressed form of historical data. */

#if 0

write_side_statistics(side, fp)
Side *side;
FILE *fp;
{
    int anystuff;
    int u, u2, i;

    /* The balance sheet of gains and losses. */
    fprintf(fp, "  (balance-sheet\n");
    for_all_unit_types(u) {
	anystuff = FALSE;
	for (i = 0; i < NUMREASONS; ++i) {
	    if (side_balance(side, u, i) > 0) { anystuff = TRUE; break; }
	}
	if (anystuff) {
	    fprintf(fp, "    (%s", u_type_name(u));
	    for (i = 0; i < NUMREASONS; ++i) {
		fprintf(fp, " %d", side_balance(side, u, i));
	    }
	    fprintf(fp, ")\n");
	}
    }
    fprintf(fp, "  )\n");
    /* The total number of attacks. */
    fprintf(fp, "  (attack-totals\n");
    for_all_unit_types(u) {
	anystuff = FALSE;
	for_all_unit_types(u2) {
	    if (side_atkstats(side, u, u2) > 0) { anystuff = TRUE; break; }
	}
	if (anystuff) {
	    fprintf(fp, "    (%s", u_type_name(u));
	    for_all_unit_types(u2) {
		fprintf(fp, " %d", side_atkstats(side, u, u2));
	    }
	    fprintf(fp, "    )\n");
	}
    }
    fprintf(fp, "  )\n");
    /* The total number of hits inflicted. */
    fprintf(fp, "  (hit-totals\n");
    for_all_unit_types(u) {
	anystuff = FALSE;
	for_all_unit_types(u2) {
	    if (side_hitstats(side, u, u2) > 0) { anystuff = TRUE; break; }
	}
	if (anystuff) {
	    fprintf(fp, "    (%s", u_type_name(u));
	    for_all_unit_types(u2) {
		fprintf(fp, " %d", side_hitstats(side, u, u2));
	    }
	    fprintf(fp, "    )\n");
	}
    }
    fprintf(fp, "  )\n");
}
#endif



reshaped_point(x1, y1, x2p, y2p)
int x1, y1, *x2p, *y2p;
{
    *x2p = (((x1 - reshaper->subareax) * reshaper->finalsubareawidth )
	    / reshaper->subareawidth ) + reshaper->finalsubareax;
    *y2p = (((y1 - reshaper->subareay) * reshaper->finalsubareaheight)
	    / reshaper->subareaheight) + reshaper->finalsubareay;
    return TRUE;
}

original_point(x1, y1, x2p, y2p)
int x1, y1, *x2p, *y2p;
{
    *x2p = (((x1 - reshaper->finalsubareax) * reshaper->subareawidth )
	    / reshaper->finalsubareawidth ) + reshaper->subareax;
    *y2p = (((y1 - reshaper->finalsubareay) * reshaper->subareaheight)
	    / reshaper->finalsubareaheight) + reshaper->subareay;
    return inside_area(*x2p, *y2p);
}

