/* Copyright (c) 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

/* Code shared between xconq and xtconq. */

#include "../src/conq.h"
#include "../xlib/xconq.h"

extern char *bitmaplib;

/* various bitmap definitions. */

#include "snowy"

#include "lookglass"
#include "lookmask"
#include "movecursor"
#include "movemask"
#include "shootcursor"
#include "shootmask"
/* need masks for these too */
#include "hexcursor"
#include "bordcursor"
#include "conncursor"
#include "addunits"

#include "boxcurs"
#include "boxmask"

#include "bomb1"
#include "bomb2"
#include "bomb3"
#include "bomb4"

#include "miss"
#include "hit"
#include "kill"

#include "hex8"
#include "hex8b"

#include "hex16"
#include "hex16b"

#include "closer"
#include "farther"
#include "terr"
#include "elev"
#include "therm"
#include "units"
#include "names"
#include "people"
#include "resource"

/* This is the basic cell size. */

int mags[] = { 1, 2, 4, 8, 16, 32, 64, 128 };

/* These give the total dimensions of each hex cell, plus the vertical
   distance center-to-center.  This is all carefully calculated to make
   the cells overlap perfectly at each different magnification, assuming
   that the icons have the right shape and size. */

int hws[] = { 1, 2, 4, 10, 20, 40, 80, 160 };
int hhs[] = { 1, 2, 5, 11, 22, 44, 88, 176 };
int hcs[] = { 1, 2, 4,  9, 17, 34, 68, 136 };

/* The sizes of the unit subcells.  This is available drawing area, exact
   unit icon sizes depends on what's available. */

/* Adjust uh down to fix exactly within a hex. */
/* (also mention drawing size sep from allowed unit icon sizes) */

int uhs[]  = { 1, 2, 2,  8, 16, 32, 64, 128 };
int uhis[] = { 1, 2, 2,  8, 14, 28, 64, 128 };
int uws[]  = { 1, 2, 2,  8, 16, 32, 64, 128 };

/* (add arrays of info about various drawing styles and allowable options) */

enum whattouse terrstyles[] =
  { useblocks, useblocks, useblocks, useblocks,
    usepolygons, usepolygons, usepolygons, usepolygons };

/* Widths of borders and connections (0 implies don't draw at all). */

int bwid[] = { 0, 0, 0, 1, 3, 5, 7, 9 };

int cwid[] = { 0, 0, 0, 1, 3, 5, 7, 9 };

/* Coordinates of the hex borders. */
/* Note that array has extra entry so don't need to wrap index. */

int bsx[NUMPOWERS][7] = {
    { 0 },
    { 0 },
    { 0 },
    {  5,  10,  10,  5,  0,  0,  5 },
    { 10,  20,  20, 10,  0,  0, 10 },
    { 20,  40,  40, 20,  0,  0, 20 },
    { 40,  80,  80, 40,  0,  0, 40 },
    { 80, 160, 160, 80,  0,  0, 80 }
};
int bsy[NUMPOWERS][7] = {
    { 0 },
    { 0 },
    { 0 },
    {  0,  2,  9, 11,  9,  2,  0 },
    {  0,  4, 18, 22, 18,  4,  0 },
    {  0,  8, 36, 44, 36,  8,  0 },
    {  0, 16, 72, 88, 72, 16,  0 },
    {  0, 32, 144, 176, 144, 32,  0 }
};

/* Coords of middles of each hex border (half a connection, basically). */
 
int lsx[NUMPOWERS][6] = {
    { 0 },
    { 0 },
    { 0 },
    {  2,  5,   2,  -2,  -5,  -2 },
    {  5, 10,   5,  -5, -10,  -5 },
    { 10, 20,  10, -10, -20, -10 },
    { 20, 40,  20, -20, -40, -20 },
    { 40, 80,  40, -40, -80, -40 }
};
int lsy[NUMPOWERS][6] = {
    { 0 },
    { 0 },
    { 0 },
    {  -4,  0,   4,   4,   0,  -4 },
    {  -9,  0,   9,   9,   0,  -9 },
    { -18,  0,  18,  18,   0, -18 },
    { -36,  0,  36,  36,   0, -36 },
    { -72,  0,  72,  72,   0, -72 }
};


init_bitmaps(side)
Side *side;
{
    Display *dpy = side->ui->dpy;
    Window win = side->ui->rootwin;

    /* Get the solid hex outlines. */
    side->ui->hexpics[3] =
      XCreateBitmapFromData(dpy, win, hex8_bits, hex8_width, hex8_height);
    side->ui->bhexpics[3] =
      XCreateBitmapFromData(dpy, win, hex8b_bits, hex8b_width, hex8b_height);
    side->ui->hexpics[4] =
      XCreateBitmapFromData(dpy, win, hex16_bits, hex16_width, hex16_height);
    side->ui->bhexpics[4] =
      XCreateBitmapFromData(dpy, win, hex16b_bits, hex16b_width, hex16b_height);
    /* Get pictures of mushroom clouds. */
    side->ui->bombpics[0] =
      XCreateBitmapFromData(dpy, win, bomb1_bits, bomb1_width, bomb1_height);
    side->ui->bombpics[1] =
      XCreateBitmapFromData(dpy, win, bomb2_bits, bomb2_width, bomb2_height);
    side->ui->bombpics[2] =
      XCreateBitmapFromData(dpy, win, bomb3_bits, bomb3_width, bomb3_height);
    side->ui->bombpics[3] =
      XCreateBitmapFromData(dpy, win, bomb4_bits, bomb4_width, bomb4_height);
    side->ui->hitpics[0] =
      XCreateBitmapFromData(dpy, win, miss_bits, miss_width, miss_height);
    side->ui->hitpics[1] =
      XCreateBitmapFromData(dpy, win, hit_bits, hit_width, hit_height);
    side->ui->hitpics[2] =
      XCreateBitmapFromData(dpy, win, kill_bits, kill_width, kill_height);
    side->ui->boxcurs =
      XCreateBitmapFromData(dpy, win, boxcurs_bits, boxcurs_width, boxcurs_height);
    side->ui->boxmask =
      XCreateBitmapFromData(dpy, win, boxmask_bits, boxmask_width, boxmask_height);
    DGprintf("Bitmaps stored ...\n");
}

/* Make all the different kinds of cursors we intend to use.  Should be
   one for each kind of mode or tool. */

init_cursors(side)
Side *side;
{
    int i, w, h;
    Display *dpy = side->ui->dpy;
    Window win = side->ui->rootwin;
    int fg = side->ui->fgcolor, bg = side->ui->bgcolor;

    for (i = 0; i < NUMCONTROLS; ++i) {
	side->ui->toolpics[i] = None;
    }
    for (i = 0; i < NUMTOOLS; ++i) {
	side->ui->toolcursors[i] = None;
    }
    rootcursor = FALSE;
    XQueryBestCursor(dpy, win, lookglass_width, lookglass_height, &w, &h);
    DGprintf("Best cursor shape is %dx%d\n", w, h);
    if (w >= lookglass_width && h >= lookglass_height) {
        side->ui->toolcursors[LOOKTOOL] =
	  make_cursor(dpy, win,
		      lookglass_bits, lookmask_bits, fg,
		      bg, lookglass_x_hot, lookglass_y_hot);
	side->ui->toolpics[LOOK] =
	  XCreateBitmapFromData(dpy, win, lookglass_bits, 16, 16);
        side->ui->toolcursors[UNITMOVETOOL] =
	  make_cursor(dpy, win,
		      movecursor_bits, movemask_bits, fg,
		      bg, movecursor_x_hot, movecursor_y_hot);
	side->ui->toolpics[UNITMOVE] =
	  XCreateBitmapFromData(dpy, win, movecursor_bits, 16, 16);
        side->ui->toolcursors[UNITSHOOTTOOL] =
	  make_cursor(dpy, win,
		      shootcursor_bits, shootmask_bits, fg,
		      bg, shootcursor_x_hot, shootcursor_y_hot);
	side->ui->toolpics[UNITSHOOT] =
	  XCreateBitmapFromData(dpy, win, shootcursor_bits, 16, 16);

	side->ui->toolpics[ZOOMOUT] =
	  XCreateBitmapFromData(dpy, win, farther_bits, 16, 16);
	side->ui->toolpics[ZOOMIN] =
	  XCreateBitmapFromData(dpy, win, closer_bits, 16, 16);
#ifdef DESIGNERS
	/* (should collect separate bitmaps for subtypes, switch between) */
        side->ui->toolcursors[HEXPAINTTOOL] =
	  make_cursor(dpy, win,
		      hexcursor_bits, hexcursor_bits, fg,
		      bg, hexcursor_x_hot, hexcursor_y_hot);
	side->ui->toolpics[HEXPAINT] =
	  XCreateBitmapFromData(dpy, win, hexcursor_bits, 16, 16);
        side->ui->toolcursors[UNITADDTOOL] =
	  make_cursor(dpy, win,
		      addunits_bits, addunits_bits, fg,
		      bg, addunits_x_hot, addunits_y_hot);
	side->ui->toolpics[UNITADD] =
	  XCreateBitmapFromData(dpy, win, addunits_bits, 16, 16);
#endif
    } else {
	init_warning("can't have %dx%d cursors on \"%s\", using default",
		     lookglass_width, lookglass_height, "" /* dpyname */);
        rootcursor = TRUE;
    }
    DGprintf("Cursors stored ...\n");
}

/* Since XCreatePixmapCursor() takes XColors and not pixel values we
   have to look up the colors associated with the foreground and
   background pixel values in the color table and pass them to
   XCreatePixmapCursor(). */
   
Cursor
make_cursor(dpy, win, cursbits, maskbits, fg, bg, x, y)
Display *dpy;
Window win;
char *cursbits, *maskbits;
unsigned long fg, bg;
unsigned int x, y;
{
    Pixmap curs, mask;
    XColor fgcolor, bgcolor;
    Cursor rslt;

    mask = XCreateBitmapFromData(dpy, win, maskbits, 16, 16);
    curs = XCreateBitmapFromData(dpy, win, cursbits, 16, 16);
    fgcolor.pixel = fg;
    XQueryColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)), &fgcolor);
    bgcolor.pixel = bg;
    XQueryColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)), &bgcolor);
    rslt = XCreatePixmapCursor(dpy, curs, mask, &fgcolor, &bgcolor, x, y);
    XFreePixmap(dpy, mask);
    XFreePixmap(dpy, curs);
    return rslt;
}

init_unit_images(side)
Side *side;
{
    int p, u, i;
    ImageFamily *uimf;
    Image *uimg;

    side->ui->uimages =
      (ImageFamily *) xmalloc(numutypes * sizeof(ImageFamily));
    for_all_unit_types(u) {
	uimf = &(side->ui->uimages[u]);
	init_image_family(side, uimf);
	if (u_image_name(u) && (strlen(u_image_name(u)) > 0)) {
	    uimf->name = u_image_name(u);
	} else {
	    uimf->name = u_type_name(u);
	}
	load_image_family(side, uimf);
	if (uimf->numsizes == 0) {
	    /* Build a default unit image. */
	    uimg = &(uimf->images[uimf->numsizes++]);
	    uimg->w = 20;  uimg->h = 22;
	    uimg->minw = 1;  uimg->minh = 1;
	    uimg->maxw = 999;  uimg->maxh = 999;
	    uimg->mono = 
	      XCreateBitmapFromData(side->ui->dpy, side->ui->rootwin,
				    hex8b_bits, hex8b_width, hex8b_height);
	}
	if (DebugG) {
	    describe_image_family(side, "unit type", u_type_name(u), uimf);
	}
    }
}

/* Collect images for all the terrain types. */

init_terrain_images(side)
Side *side;
{
    int i, t, p;
    int junk1, junk2;
    int foundsome = FALSE;
    short w, h;
    Pixmap pat;
    int w2, h2;
    ImageFamily *timf;
    Image *timg;

    side->ui->timages =
      (ImageFamily *) xmalloc(numttypes * sizeof(ImageFamily));
    for_all_terrain_types(t) {
	timf = &(side->ui->timages[t]);
	init_image_family(side, timf);
	if (t_image_name(t) && (strlen(t_image_name(t)) > 0)) {
	    timf->name = t_image_name(t);
	} else {
	    timf->name = t_type_name(t);
	}
	load_image_family(side, timf);
	if (timf->numsizes == 0) {
	    /* Build a default terrain image. */
	    timg = &(timf->images[timf->numsizes++]);
	    timg->w = 20;  timg->h = 22;
	    timg->minw = 1;  timg->minh = 1;
	    timg->maxw = 999;  timg->maxh = 999;
	    timg->mono = 
	      XCreateBitmapFromData(side->ui->dpy, side->ui->rootwin,
				    hex16b_bits, hex16b_width, hex16b_height);
	}
	if (DebugG) {
	    describe_image_family(side, "terrain type", t_type_name(t), timf);
	}
    }
}

describe_image_family(side, classname, typename, imf)
Side *side;
char *classname, *typename;
ImageFamily *imf;
{
    int i;
    Image *img;

    DGprintf("%s %s family for %s has %d images\n",
	     classname, typename, side_desig(side), imf->numsizes);
    for (i = 0; i < imf->numsizes; ++i) {
	img = &(imf->images[i]);
	DGprintf("    %dx%d\n", img->w, img->h);
    }
}

init_image_family(side, imf)
Side *side;
ImageFamily *imf;
{
    bzero(imf, sizeof(ImageFamily));
}

Image *
get_img(imf, w, h)
ImageFamily *imf;
int w, h;
{
    int i;
    Image *img;

    for (i = 0; i < imf->numsizes; ++i) {
	img = &(imf->images[i]);
	if (w == img->w && h == img->h) return img;
    }
    if (imf->numsizes >= 10 /* MAXIMSIZES */) return NULL;
    img = &(imf->images[imf->numsizes]);
    ++(imf->numsizes);
    img->w = w;  img->h = h;
    /* Default min/max limits to actual size. */
    img->minw = img->maxw = w;  img->minh = img->maxh = h;
    return img;
}

/* This tries to fill in the given image family. */

load_image_family(side, imf)
Side *side;
ImageFamily *imf;
{
    int w, h, i;
    Pixmap pic;
    Image *img;

    /* If no name, don't even try. */
    if (imf->name == NULL) return;
    if (0 /* found an imf file */) {
    } else {
	/* Just grab up plausibly-named bitmaps. */
	if ((pic = load_bitmap(side, imf->name, "b", &w, &h)) != None) {
	    if ((img = get_img(imf, w, h)) != NULL) {
		img->mono = pic;
	    }
	}
	if ((pic = load_bitmap(side, imf->name, "32.b", &w, &h)) != None) {
	    if ((img = get_img(imf, w, h)) != NULL) {
		img->mono = pic;
	    }
	}
	if ((pic = load_bitmap(side, imf->name, "16.t", &w, &h)) != None) {
	    if ((img = get_img(imf, w, h)) != NULL) {
		img->mono = pic;
	    }
	}
	if ((pic = load_bitmap(side, imf->name, "t", &w, &h)) != None) {
	    if ((img = get_img(imf, w, h)) != NULL) {
		img->mono = pic;
		img->mask = pic;
	    }
	}
	if ((pic = load_bitmap(side, imf->name, "e", &w, &h)) != None) {
	    if ((img = get_img(imf, w, h)) != NULL) {
		img->mono = pic;
		img->mask = pic;
	    }
	}
	if (imf->numsizes > 0) {
	    return TRUE;
	} else {
	    fprintf(stderr,
		    "Image named \"%s\" not found here or in \"%s\"!\n",
		    imf->name, bitmaplib);
	}
    }
#if 0
    /* Now attempt to pick up any fonts that might be available. */
    /* (How to do more than one?) */
    side->ui->iconfont = open_font(side, ICONFONT, "IconFont", NULL, "");
    if (side->ui->iconfont != NULL) {
	/* See if the font matches any of the magnifications. */
	w = font_width(side->ui->iconfont);
	h = font_height(side->ui->iconfont);
	for (p = 0; p < NUMPOWERS; ++p) {
	    if (hws[p] == w && hhs[p] == h) {
		for_all_terrain_types(t) {
		    /* should also test for presence of each terrtype */
		    side->ui->usewhat[p][t] = usefontchars;
		    side->ui->terrchars[p][t] = t_type_char(t);
		    side->ui->terrfonts[p][t] = side->ui->iconfont;
		    foundsome = TRUE;
		}
	    }
	}
	if (!foundsome) {
	    init_warning("The font \"%s\" (size %dx%d) was useless",
			 ICONFONT, w, h);
	}
    } else {
	/* Be informative, especially for new X users. */
	/* (change to identify the specific fonts that should be installed?) */
	init_warning("Xconq is usually much faster if fonts are installed.");
	init_warning("Please consult the docs and/or your site admins for details.");
	/* Mark this as undefined. */
	side->ui->iconfont = NULL;
    }
#endif
#if 0
    /* (Should be able to skip font loading if units don't need it) */
    side->ui->unitfont = open_font(side, "standard", "UnitFont",
				    side->ui->textfont);
    sw = font_width(side->ui->unitfont);
    sh = font_height(side->ui->unitfont);
    /* shouldn't set any of this up unless it knows that the font & char
       are the right ones, at whatever power. */
    /* See if the font is usable at any of the magnifications. */
    for (p = 0; p < NUMPOWERS; ++p) {
	/* OK to use small icons on big hexes. */
	if (uws[p] >= sw && uhs[p] >= sh) {
	    for_all_unit_types(u) {
		side->ui->unitchars[p][u] = u_type_char(u);
		side->ui->unitfonts[p][u] = side->ui->unitfont;
		side->ui->unitw[p][u] = sw;  side->ui->unith[p][u] = sh;
		foundsome = TRUE;
	    }
	}
    }
    if (!foundsome) {
	init_warning("The font \"%s\" (size %dx%d) was useless",
		     "standard", sw, sh);
    }
#endif
}

/* Try to load a bitmap of the given name, looking in both the current dir
   and the library dir.  Warns about failure, but returns None, so the rest
   of the code will continue to function. */

Pixmap
load_bitmap(side, name, ext, wp, hp)
Side *side;
char *name, *ext;
int *wp, *hp;
{
    int hotx, hoty;
    unsigned int w, h;
    Pixmap rslt;

    if (ext != NULL) {
	make_pathname(NULL, name, ext, spbuf);
	if (XReadBitmapFile(side->ui->dpy, side->ui->rootwin, spbuf,
			    &w, &h, &rslt, &hotx, &hoty) == BitmapSuccess) {
	    DGprintf("Loaded bitmap \"%s\"\n", spbuf);
	    *wp = w;  *hp = h;
	    return rslt;
	}
	make_pathname(bitmaplib, name, ext, spbuf);
	if (XReadBitmapFile(side->ui->dpy, side->ui->rootwin, spbuf,
			    &w, &h, &rslt, &hotx, &hoty) == BitmapSuccess) {
	    DGprintf("Loaded bitmap \"%s\"\n", spbuf);
	    *wp = w;  *hp = h;
	    return rslt;
	}
    }
    return None;
}
