/* Inspired by ls, based on xdb by David Curry, davy@riacs.edu. */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
/*#include <X11/Xaw/List.h> */
#include <X11/Xaw/Command.h>
#include <X11/Shell.h>

#include "xtbmlist.h"

#include <stdio.h>

#define MAXFILES 1000

char *pname = "xlsb";

char *bitmapfiles[MAXFILES];
int nfiles = 0;

int rows, cols;
int colwid = 100;
int dpywid;

int drawbbox = 0;

Arg tmpargs[10];

    Widget toplevel, mainbox, tmpwidget, thelist;
    XtAppContext app;

void
shut_down(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    exit(0);
}

void
toggle_frame(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, flag;

    i = 0;
    XtSetArg(tmpargs[i], XtNframe, &flag);  ++i;
    XtGetValues(thelist, tmpargs, i);
    flag = !flag;
    i = 0;
    XtSetArg(tmpargs[i], XtNframe, flag);  ++i;
    XtSetValues(thelist, tmpargs, i);
}

main(argc, argv)
int argc;
char **argv;
{
    int i;

    toplevel = XtAppInitialize(&app, "xlsb",
			       (XrmOptionDescList) NULL, 0, &argc, argv,
			       (String *) NULL, (ArgList) NULL, 0);

    while (--argc) {
	++argv;
	if (!strcmp(*argv, "-bb")) {
	    drawbbox = 1;
	    continue;
	}
	if (!strcmp(*argv, "-frame")) {
	    drawbbox = 1;
	    continue;
	}
	if (!strcmp(*argv, "-px")) {
	    /* set bitmap pixel size */
	    ++argv;
	    continue;
	}
	if (!strcmp(*argv, "-size")) {
	    /* pick size or range of sizes to show */
	    ++argv;
	    continue;
	}
	if (!strcmp(*argv, "-bysize")) {
	    /* sort by sizes first */
	    ++argv;
	    continue;
	}
	if (!strcmp(*argv, "-mask")) {
/*	    strcpy(maskext, *(++argv));  */
	    continue;
	}
	if (*argv != NULL) {
	    bitmapfiles[nfiles++] = *argv;
	}
    }
    /* Must have at least one file (could dig through current dir?). */
    if (nfiles == 0) {
	fprintf(stderr, "No bitmaps specified!\n");
	exit(1);
    }
    mainbox = XtCreateManagedWidget("box", boxWidgetClass,
				toplevel, (Arg *) NULL, 0);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Frame");  ++i;
    tmpwidget =
      XtCreateManagedWidget("frame", commandWidgetClass,
			    mainbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, toggle_frame, NULL);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Quit");  ++i;
    tmpwidget =
      XtCreateManagedWidget("quit", commandWidgetClass,
			    mainbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, shut_down, NULL);

    dpywid = 1000; /* or some such */
    colwid = 200; /* but should calc from images/string lengths */
    cols = dpywid / colwid;
    /* should adjust cols up/down for extreme cases */
    rows = nfiles / cols + 1;
    cols = nfiles / rows + 1;

    i = 0;
    XtSetArg(tmpargs[i], XtNlist, bitmapfiles);  ++i;
    XtSetArg(tmpargs[i], XtNnumberStrings, nfiles);  ++i;
    XtSetArg(tmpargs[i], XtNverticalList, True); ++i;
    XtSetArg(tmpargs[i], XtNrowSpacing, 20); ++i;
    XtSetArg(tmpargs[i], XtNdefaultColumns, cols); ++i;
    XtSetArg(tmpargs[i], XtNframe, drawbbox); ++i;
    thelist =
      XtCreateManagedWidget("bitmaplist", bitmapListWidgetClass,
			    mainbox, tmpargs, i);

    XtRealizeWidget(toplevel);

    XtAppMainLoop(app);
}

#if 0
/* Read all the bitmaps and compute their desired layout. */

read_and_layout(widget)
Widget widget;
{
#if 0
    Pixmap read_bitmap_file();
    unsigned int x, y, width, height;
    int n;

#if 0
    font = XLoadQueryFont(dpy, "fixed");
#endif
    fontwid = 10 /* font->max_bounds.width */;
    fonthgt = 15;

    /* Get all the bitmaps and compute sizes. */
    for (n = 0; n < nfiles; ++n) {
	int fullwid;

	bitmaps[n] = read_bitmap_file(bitmapfiles[n], &width, &height);
	if (bitmaps[n] != None) {
	    ws[n] = width;  hs[n] = height;
	    if (width > maxwid) maxwid = width;
	    if (height > maxhgt) maxhgt = height;
	    fullwid = width + 2 + fontwid * (strlen(bitmapfiles[n]) + 2);
	    if (fullwid > colwid) colwid = fullwid;
	}
    }
    /* Decide on a tasteful layout of the bitmaps. */
    dpywid = 1000 /* XDisplayWidth(dpy, screen)*/;
    if (dpywid > 1000) dpywid = 1000;
    cols = dpywid / colwid;
    /* should adjust cols up/down for extreme cases */
    rows = nfiles / cols + 1;
    cols = nfiles / rows + 1;
    /* This doesn't seem to work... */
    n = 0;
    XtSetArg(resizeargs[n], XtNwidth, cols * colwid);
    ++n;
    XtSetArg(resizeargs[n], XtNheight, rows * (maxhgt + 4));
    ++n;
    XtSetValues(widget, resizeargs, n);
    /* Good a time as any to create the one gc we'll need. */
    /* Can't be shared because we'll be messing with it. */
    {
	XGCValues values;

	values.foreground = 0;
/*	values.font = ?;  */
	gc = XCreateGC(dpy, lswin, GCForeground, &values);
    }
#endif
}

/* Blat a single bitmap+mask+bbox_name onto the window. */

draw_one_bitmap(win, bitmap, name, x, y, width, height)
Window win;
Pixmap bitmap;
char *name;
int x, y;
unsigned int width, height;
{
#if 0
    char *nameproper, *rindex();

    if (bitmap == None) return;

    /* Draw the bounding box if it's been asked for. */
    if (drawbbox) {
	XSetClipMask(dpy, gc, None);
	XSetForeground(dpy, gc, 0);
	XFillRectangle(dpy, win, gc,
		   x + (maxwid - width) / 2 , y + (maxhgt - height) / 2,
		   width, height);
	XSetForeground(dpy, gc, 1);
    }	
    /* Draw the bitmap. */
    XSetClipMask(dpy, gc, bitmap);
    XSetClipOrigin(dpy, gc,
		   x + (maxwid - width) / 2 , y + (maxhgt - height) / 2);
    XFillRectangle(dpy, win, gc,
		   x + (maxwid - width) / 2 , y + (maxhgt - height) / 2,
		   width, height);
    /* Extract the minimal name. */
    if ((nameproper = rindex(name, '/')) == NULL)
      nameproper = name;
    else
      nameproper++;
    /* Draw the name of the bitmap. */
    XSetClipMask(dpy, gc, None);
    /* (need to acquire the right font for this gc) */
    XDrawImageString(dpy, win, lsgc, x + maxwid + 1, y + maxhgt / 2,
		     nameproper, strlen(nameproper));
#endif
}

usage()
{
    fprintf(stderr, "%s <usual X args> ", pname);
    fprintf(stderr, "[-bb] ... bitmapfiles ...\n");
    exit(1);
}
#endif
