/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

/* Interface definitions for the Xlib version of xconq. */

/* Default color of text and icons - 0 is for white on black, 1 */
/* is for black on white.  Should be set appropriately for the most */
/* common monochrome display (color displays always do white on black). */
/* This is also settable by the player, so the default is just for */
/* convenience of the majority. */

#define BLACKONWHITE 1

#define BARGRAPHS

/* The default fonts can be altered by users, so these are just hints. */

#define TEXTFONT "fixed"
#define HELPFONT "fixed"
#define ICONFONT "xconq"

#define STANDARD "standard"


/* Number of messages displayed at one time. No upper limit I believe, */
/* but too many won't fit on the screen.  The actual numbers of lines */
/* displayed can be changed by the player, subject to limitations on the */
/* screen space available. */
/* (make this totally dynamic?) */

#define MAXNOTES 45

/* All names, phrases, and messages must be able to fit in statically */
/* allocated buffers of this size. */
/* (shouldn't these be flushed?) */

#define BUFSIZE 120
#define NOTIFY_SPACE  MAXNOTES * 60 * MAXSIDES + BUFSIZE

/* Some X11 servers die if too much is written between output flushes. */

/* #define STUPIDFLUSH */

/* This is the name of a family of programs, so argv[0] inadequate. */

#define PROGRAMNAME "xconq"

#define BITMAPLIB "../pics"

/* Be careful of the path of the X header files. */

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>

/* should find a better home? */

/* How many magnifications we're supporting. */

#define NUMPOWERS 8

#include "../x11/xcommon.h"

#include "ximf.h"

/* Magic values (unlikely to show up as player input) */

#define DFLTFLAG (-12345)       /* flags when default arg to be used */
#define ITERFLAG (-12347)       /* flags args that are "itertime" */

#define ESCAPE '\033'           /* standard abort character */
#define BACKSPACE '\010'        /* for fixing strings being entered */

/* The list of defined controls. */

enum controltypeenum {
    LOOK = 0,  LOOKAUX,
    UNITMOVE,  UNITMOVEAUX,
    UNITSHOOT, UNITSHOOTAUX,
    UNITBUILD, UNITBUILDAUX,
    HOLE18,    HOLE19,
    SHOWTERR,  HOLE, 
    SHOWUNITS, HOLE93, 
    SHOWPATS,  HOLE94, 
    SHOWNAMES, HOLE95, 
    SHOWEDGES, HOLE96, 
    SHOWPEOPLE,HOLE97, 
    SHOWELEVS, HOLE98, 
    SHOWTEMP,  HOLE99,
    ZOOMOUT,   ZOOMIN,
    /* need more display controls here */
    HOLE14,    HOLE15,
    HEXPAINT,  CURTTYPE,
    UNITADD,   CURUTYPE,
    CURSIDE,   HOLExx,
    numcontrols
};

typedef struct a_controltype {
    enum controltypeenum type;
    char *name;
    int (*fn)();
    int tool;
    int fordesigners;
    int x, y, w, h;
} Controltype;

extern Controltype controltypes[];

/* This value should be larger than "numcontrols". */

#define NUMCONTROLS 40

/* The dynamic state of control panel items. */

typedef struct a_control {
    int visible;        /* true if being displayed at all */
    int enabled;        /* true if can click on this */
    int selected;       /* true if currently being used */
    char *text;          /* text to display */
    long color;          /* what color to use */
    Pixmap pic;          /* what image to use here */
    int (*drawer)();     /* arb fn to draw in control */
} Control;

enum whattouse {
  dontdraw,
  useblocks,
  usepictures,
  usefontchars,
  usepolygons
};

/* Tools govern the interpretation of mouse clicks in a map view. */

enum tooltypes {
    LOOKTOOL = 0,
    UNITMOVETOOL = 1,
    UNITSHOOTTOOL = 2,
    HEXPAINTTOOL = 3,
    UNITADDTOOL = 4,
    NUMTOOLS = 5
};

/* Bounding box size for bomb images. */

#define BH 32

/* The user interface substructure.  This is only allocated for sides with
   attached displays. */

typedef struct a_ui {
    Display *dpy;             /* The X display used by this side */
    int screen;               /* The X screen in use */
    Window rootwin;           /* The root window */

    short bonw;               /* true if display is black-on-white */
    short followaction;       /* move to where a change has occured */

    /* input hacking */
    struct a_map *inpmap;     /* which map all this was on */
    struct a_map *downmap;    /* which map all this was on */
    int downx, downy, downdir;   /* press map coordinates and hexside */
    struct a_map *upmap;      /* which map all this was on */
    int upx, upy, updir;      /* release map coordinates and hexside */
    /* stuff for modal dialogs */
    int (*modalhandler)();    /* function that handles modal input */

    struct a_unit *curunit;

    /* Random input data slots - flush these! */
    char reqch;               /* keyboard char */
    struct a_side *reqoside;
    struct a_unit *requnit;
    int reqvalue, reqvalue2;
    int reqposx, reqposy;     /* accumulator for position reading */
    char ustr[MAXUTYPES];     /* used in composing unit type hints */
    int uvec[MAXUTYPES];      /* vector of allowed unit types to input */
    int bvec[MAXUTYPES];      /* bit vector of allowed unit types to input */

    char prompt[BUFSIZE];
    char answer[BUFSIZE];

    /* Constructed during display init */
    short monochrome;         /* obvious */
    short fw, fh;             /* dimensions of text font (in pixels) */
    /* Working variables for the display */
    short lastx, lasty;       /* last current x and y (-1,-1 initially) */
    /* All the colors used - filled in by display init */
    long bgcolor;             /* background color */
    long fgcolor;             /* foreground (text) color */
    long owncolor;            /* color for us (usually black) */
    long enemycolor;          /* color for them (usually red) */
    long neutcolor;           /* color for fencesitters (usually gray) */
    long graycolor;           /* color for graying out (usually gray) */
    long diffcolor;           /* unusual/distinct color (usually maroon) */
    long goodcolor;           /* color for OKness (usually green) */
    long badcolor;            /* color for non-OKness (usually red) */
    long hexcolor[MAXTTYPES]; /* the color of each terrain type */
    long hexfgcolor[MAXTTYPES]; /* the "fg" color of each terrain type */
    long ourcolors[3];        /* national colors */
    long ournumcolors;
    long ouremblem;          /* pattern for national emblem (pixmap?) */
    long colors[MAXSIDES][3];
    int numcolors[MAXSIDES];
    struct a_map *maps;       /* chain of maps that we're using */
    struct a_map *front;      /* the map currently "in use" */
    Window help;                /* help window */
    Window helpinfo;            /* help content subwindow */

    struct a_helpnode *curhelpnode;
    int active;
    GC gc;			/* a tmp graphics context for this display */
    GC textgc;			/* foreground on background text */
    GC ltextgc;			/* foreground on background text */
    GC icongc;			/* icon graphics context */
    GC invicongc;		/* icon gc inverted colors. */
    GC varicongc;		/* This is an icon gc with variable fgcolor. */
    GC unitgc;			/* unit icon display gc. */
    GC cleargc;			/* The gc to use for clearing areas. */
    GC clearbitgc;              /* gc for clearing bitmaps */
    GC helpgc;
    XFontStruct *textfont;	/* font for text display */
    XFontStruct *helpfont;	/* font for text display */
    /* need legend fonts for other things too */
    XFontStruct *cursorfont;      /* font for cursors */

    /* Terrain drawing machinery */
    enum whattouse usewhat[NUMPOWERS][MAXTTYPES];

    /* (should move into image family) */
    Pixmap terrpics[NUMPOWERS][MAXTTYPES]; /* terrain bitmaps */
    char terrchars[NUMPOWERS][MAXTTYPES];
    XFontStruct *terrfonts[NUMPOWERS][MAXTTYPES];
    Pixmap hexpics[NUMPOWERS], bhexpics[NUMPOWERS];
    Pixmap hexchars[NUMPOWERS], bhexchars[NUMPOWERS];
    XFontStruct *hexfonts[NUMPOWERS], *bhexfonts[NUMPOWERS];
    XFontStruct *iconfont;
    /* Unit drawing machinery */
    Pixmap unitpics[NUMPOWERS][MAXUTYPES];	/* unit bitmaps */
    Pixmap unitmasks[NUMPOWERS][MAXUTYPES];
    char unitchars[NUMPOWERS][MAXUTYPES];
    XFontStruct *unitfonts[NUMPOWERS][MAXUTYPES];
    XFontStruct *unitfont;	/* font for unit characters */
    int unitw[NUMPOWERS][MAXUTYPES], unith[NUMPOWERS][MAXUTYPES];
    /* Emblem drawing machinery */
    Pixmap emblempics[MAXSIDES];
    Pixmap emblemmasks[MAXSIDES];
    int embw[MAXSIDES], embh[MAXSIDES];
    /* Map legend drawing machinery */

    XFontStruct *ulegendfonts[NUMPOWERS][MAXUTYPES];
    XFontStruct *flegendfonts[NUMPOWERS][10];  /* num feature types? */

    /* Random stuff */
    Pixmap bombpics[4];		/* mushroom clouds */
    Pixmap hitpics[3];
    Pixmap toolpics[NUMCONTROLS];
    Pixmap toolcursors[NUMTOOLS];
    Pixmap boxcurs, boxmask;

    ImageFamily *uimages;
    ImageFamily *timages;
    ImageFamily *eimages;

    int numlowcontrast;
    XColor low1[50];
    XColor low2[50];

} UI;

/* Each side can open up any number and shape of maps. */

/* All the maps will share colors and bitmaps though. */

typedef struct a_map {
    short totalw, totalh;     /* Total size of window in pixels */
    short leftfrac;
    short leftw;
    short pxw, pxh;           /* Size of map subwindow in pixels */
    short msgw, msgh;
    short infow, infoh;
    short listsh;
    short list1frac;
    short list1w, list1h;
    short list2w, list2h;
    short curx, cury;         /* current spot being looked at */
    struct a_unit *curunit;   /* unit under cursor */
    short vx, vy;             /* lower left of viewport in hex coords */
    short vw, vh;             /* Boundaries of viewport in hex coords */
    short vw2, vh2;           /* 1/2 (rounded down) of above values */
    short power;              /* index to this map's magnification */
    short mag;                /* magnification of each hex (a power of 2) */
    short vcx, vcy;           /* center of the view */
    short lastvcx, lastvcy;
    short hw, hh;             /* pixel dims of a hex cell */
    short hch;                /* height of hex cell between centers */
    short uw, uh;             /* pixel dims of unit subcell */
    short sidespacing;
    /* Subwindows owned by the map */
    Window window;              /* If there is a main window, this is it */
    Window infowin;             /* pointer to info window if used */
    Window viewwin;
    Window controls;
    Window hscrollbar;
    Window vscrollbar;
    Window game;                /* overall game info */
    Window sides;                /* list of sides */
    Window msg;                 /* places for notices/warnings */
    Window promptwin;
    Window list;
    Window listscrollbar;
    /* How to draw the map. */
    int drawterrain;         /* Display solid color terrain on the map? */
    int drawgrid;        /* Draw outlines around hexes? */
    int drawhexpats;         /* Draw terrain patterns? */
    int drawunits;           /* Display units on the map? */
    int drawnames;           /* Display unit names/numbers on the map? */
    short oldesttoshow;       /* the relative time of the oldest data */
    short agetofade;          /* point at which to gray out the display */
    short newesttoshow;       /* the relative time of the newest data */
    short fadeterrain;        /* fade world features as well as units? */
    short lowestlevel;        /* the lowest level of units to display */
    short highestlevel;       /* the highest level of units to display */
    int drawpeople;
    int drawelevations;
    int drawresources[MAXMTYPES];
    short drawtemp;
    short drawweather;
    /* The state of the control panel. */
    short curtool;       /* interpretation of mouse clicks */
    short curshifttool;  /* interp of shifted clicks */
    struct a_control controlbs[NUMCONTROLS];  /* the state of the controls */
#ifdef DESIGNERS
    short curutype;
    short curttype;
    short cursidenumber;
    short curfeature;
#endif
    struct a_map *supermap;    /* pointer to map that includes this one */
    /* Link to the next map. */
    struct a_map *next;
} Map;

/* Iteration over all of a side's map windows. */

#define for_all_maps(s,m)  \
  for ((m) = (s)->ui->maps; (m) != NULL; (m) = (m)->next)

/* Random function declarations. */

XFontStruct *open_font();
Pixmap load_bitmap();
Cursor make_cursor();

/* Values of parameters generally tied to fonts and the like. */

extern int margin;
extern int fudge;
extern int conwid;
extern int conhgt;
extern int sbarwid;
extern int everything;
extern int rootcursor;

extern int bwid[];
extern int mags[], hws[], hhs[], hcs[], uws[], uhs[];
extern int dolin[], bsx[8][7], bsy[8][7], lsx[8][6], lsy[8][6];

#define XGetDefault(a,b,c) NULL
