/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

#include "conq.h"
#include "xtconq.h"

Side *side_from_display();

char *bitmaplib = BITMAPLIB;

int margin = 2;
int bd = 1;

/* Dimensions of a control. */

int conwid = 48;
int conhgt = 24;

/* Fixed dimension of a scroll bar. */

int sbarwid = 16;

int rootcursor;                /* true if using parent window's cursor */
int windows_up = FALSE;         /* Try to handle initial exposes. */

/* Set up the basic user interface for a side. */

init_ui(side)
Side *side;
{
    if (side_wants_display(side)) {
        side->ui = (UI *) xmalloc(sizeof(UI));
	side->ui->bottom_note = 0;
	side->ui->followaction = 0;
	side->ui->curhelpnode = NULL;
	init_requests(side);
	side->ui->active = FALSE;
    } else {
	side->ui = NULL;
    }
}

#if 0
/* Find a unit and put it at center (or close to it) of first view.  I guess */
/* can't use put_on_screen because not enough stuff is inited yet. */

init_curxy(side)
Side *side;
{
    Unit *unit;
    Side *loop_side;
    Map *map;

    for_all_units(loop_side, unit) {
	if (unit->side == side) {
	    Dprintf("Centering maps on unit at %d,%d\n", unit->x, unit->y);
	    for_all_maps(side, map) {
		if (!map->isworld) {
		    map->vcx = unit->x;
		    map->vcy = max(unit->y, map->vw2+1);
		    map->vcy = min(map->vcy, (area.height-1) - map->vh2);
/*		    set_map_viewport(map);  */
		}
	    }
	    return;
	}
    }
    Dprintf("Centering maps at %d,%d\n", area.width/2, area.height/2);
    for_all_maps(side, map) {
	map->vcx = area.width/2;  map->vcy = area.height/2;
    }
}
#endif


/* Get all the displays initialized, but without drawing any content, since
   not all the data inits are sufficiently complete yet. */

init_x_displays()
{
    Side *side, *side2;

    init_displays();
    /* Set up everybody's view of everybody else's colors and emblems. */
    for_all_sides(side) {
	if (side_has_display(side)) {
	    for_all_sides(side2) {
		if (side_has_display(side2)) {
		    init_emblem(side, side2);
		}
	    }
	    /* Don't forget neutrals. */
	    init_emblem(side, indepside);
	}
    }
}

/* Open display, create all the windows we'll need, do misc setup things,
   and initialize some globals to out-of-range values for recognition later. */

init_display(side)
Side *side;
{
    int i, s, n;
    int p, u, t;
    int argc = 0;
    char *dpyname;
    extern XtAppContext thisapp;
    Unit *unit;

    Dprintf("Will try to open display \"%s\" ...\n", side->player->displayname);

    for (s = 0; s < MAXSIDES; ++s) {
	side->ui->emblempics[s] = None;
	side->ui->emblemmasks[s] = None;
    }
    if (side->player == NULL) abort(); /* very great badness */
    dpyname = side->player->displayname;
    /* A hack. */
    if (dpyname != NULL && strcmp("here", dpyname) == 0) {
	dpyname = getenv("DISPLAY");
    }
    side->ui->dpy =
      XtOpenDisplay(thisapp, dpyname, "XtConq", "XtConq",
		    NULL, 0, &argc, NULL);
    if (side->ui->dpy == NULL) {
	init_error("Display \"%s\" could not be opened!\n", dpyname);
	/* (should have a more reasonable response... common error) */
    }
    if (DebugG) {
	XSynchronize(side->ui->dpy, TRUE);
	DGprintf("Synching the X server.\n");
    }
    toplevel(side) = XtAppCreateShell("XtConq", "XtConq",
				      applicationShellWidgetClass,
				      side->ui->dpy, NULL, 0);
    popup_game_widget(side);
    /* Reset popup widgets. */
    side->ui->utypedialog = NULL;
    side->ui->active = TRUE;
    side->ui->nw = min(BUFSIZE, 80);
    side->ui->nh = max(1, min(10, MAXNOTES));
    /* Create the generic windows. */
    /* (Use prefs to decide about default creation?) */
    side->ui->msg = 0;
    side->ui->maps = NULL;
    side->ui->front = NULL;
    side->ui->world = NULL;
    side->ui->lists = NULL;
    side->ui->frontlist = NULL;
    side->ui->closeups = NULL;
    side->ui->frontcloseup = NULL;
    side->ui->sidecloseups = NULL;
    side->ui->ourcloseup = NULL;
    for (i = 0; i < MAXNOTES; ++i) side->ui->noticebuf[i] = " ";
    for (i = 0; i < MAXUTYPES; ++i) side->ui->bvec[i] = 0;
    /* Flag some values as uninitialized */
    for (s = 0; s < MAXSIDES; ++s) {
	side->ui->emblems[s] = 0;
    }
    /* Create the "main" map first. */
    create_map(side, FALSE);
    if (0 /* preference */) create_map(side, TRUE);
    if (0 /* preference */) create_list(side, 1);
    /* Create help window but leave it unmapped (why?) */
    create_help_window(side);
    Dprintf("Successfully opened \"%s\"!\n", dpyname);
}

void
right_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, vcx;
    Map *map = (Map *) clientdata;

    i = 0;
    XtSetArg(tmpargs[i], XtNvcx, &vcx);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    vcx += 10;
    vcx = wrap(vcx);
    i = 0;
    XtSetArg(tmpargs[i], XtNvcx, vcx);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
left_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, vcx;
    Map *map = (Map *) clientdata;

    i = 0;
    XtSetArg(tmpargs[i], XtNvcx, &vcx);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    vcx -= 10;
    vcx = wrap(vcx);
    i = 0;
    XtSetArg(tmpargs[i], XtNvcx, vcx);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
up_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, vcy;
    Map *map = (Map *) clientdata;

    i = 0;
    XtSetArg(tmpargs[i], XtNvcy, &vcy);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    vcy += 10;
    i = 0;
    XtSetArg(tmpargs[i], XtNvcy, vcy);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
down_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, vcy;
    Map *map = (Map *) clientdata;

    i = 0;
    XtSetArg(tmpargs[i], XtNvcy, &vcy);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    vcy -= 10;
    i = 0;
    XtSetArg(tmpargs[i], XtNvcy, vcy);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
zoom_in_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, power;
    Side *side = side_from_display(XtDisplay(w));
    Map *map = (Map *) clientdata;

    i = 0;
    XtSetArg(tmpargs[i], XtNpower, &power);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    ++power;
    i = 0;
    XtSetArg(tmpargs[i], XtNpower, power);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
zoom_out_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, power;
    Side *side = side_from_display(XtDisplay(w));
    Map *map = (Map *) clientdata;

    i = 0;
    XtSetArg(tmpargs[i], XtNpower, &power);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    --power;
    i = 0;
    XtSetArg(tmpargs[i], XtNpower, power);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
show_terrain_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, flag;
    Side *side = side_from_display(XtDisplay(w));
    Map *map = (Map *) clientdata;

    i = 0;
    XtSetArg(tmpargs[i], XtNshowTerrain, &flag);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    flag = !flag;
    i = 0;
    XtSetArg(tmpargs[i], XtNshowTerrain, flag);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
show_units_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, flag;
    Side *side = side_from_display(XtDisplay(w));
    Map *map = (Map *) clientdata;

    i = 0;
    XtSetArg(tmpargs[i], XtNshowUnits, &flag);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    flag = !flag;
    i = 0;
    XtSetArg(tmpargs[i], XtNshowUnits, flag);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
show_names_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, flag;
    Side *side = side_from_display(XtDisplay(w));
    Map *map = (Map *) clientdata;

    map->drawnames = !map->drawnames;
    i = 0;
    XtSetArg(tmpargs[i], XtNshowNames, &flag);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    flag = !flag;
    i = 0;
    XtSetArg(tmpargs[i], XtNshowNames, flag);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
show_grid_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i, flag;
    Side *side = side_from_display(XtDisplay(w));
    Map *map = (Map *) clientdata;

    i = 0;
    XtSetArg(tmpargs[i], XtNshowGrid, &flag);  ++i;
    XtGetValues(map->viewwin, tmpargs, i);
    flag = !flag;
    i = 0;
    XtSetArg(tmpargs[i], XtNshowGrid, flag);  ++i;
    XtSetValues(map->viewwin, tmpargs, i);
}

void
cancel_map(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));
    Map *map = (Map *) clientdata;

    /* should destroy map object too somehow */
    XtPopdown(map->mapwin);
}

/* Create a generic map object. */

create_map(side, isworld)
Side *side;
int isworld;
{
    int r, sx, sy;
    Map *map = (Map *) xmalloc(sizeof(Map));
    extern int conhgt;
    int sbaradj = (map->isworld ? 0 : sbarwid);
    int i;
    XSetWindowAttributes foo;
    char *retain;
    Widget mapform, mapviewport, mapcontrols;
    Widget tmpwidget;

    Dprintf("Creating map\n");
#if 0
    map->isworld = isworld;
    if (map->isworld) {
	/* Guess at an appropriate size. */
	map->pxw = area.width * map->mag;  map->pxh = area.height * map->mag;
	if (!area.xwrap) {
	    map->pxw += (area.height / 4) * map->mag;
	}
	/* Pick an appropriate center of the view. */
	map->vcx = area.width / 2 - area.height / 4;
	map->vcy = area.height / 2;
    } else {
	/* a guess, should be cleverer */
	map->pxw = 24 * 20;  map->pxh = 24 * 17;
	map->vcx = map->vcy = -1;
    }
    /* Make sure control panel is visible (?) */
    map->pxh = max(map->pxh, 12 * conhgt);
#endif
#if 0
    /* This is designer stuff actually. */
    /* Or maybe the first linear type? */
    map->curbordtype = NOBORDER;
    map->curconntype = NOCONNECTION;
    /* Set random state/mode vars. */
    map->curttype = 0;  /* should be first can-be-hex type */
    map->curutype = 0;
    map->curunit = NULL;
    map->curside = 0;
#endif
    /* Newest map goes on the front of the list. */
    map->next = side->ui->maps;
    side->ui->maps = map;
    i = 0;
#if 0
    /* Decide on a good starting position (tries to minimize players' having
       to push the windows around after they're up) */
    if (map->isworld) {
	/* World maps down in the lower right corner. */
	sx = display_width(side) - map->pxw - 25 - 1;
	sy = display_height(side) - map->pxh - 1;
    } else {
	sx = 10;  sy = 50;
/*
	sx = display_width(side) / 2 - map->pxw / 2;
	sy = max(10, display_height(side) / 2 - map->pxh / 2);
*/
    }
    XtSetArg(tmpargs[i], XtNx, sx);  ++i;
    XtSetArg(tmpargs[i], XtNy, sy);  ++i;
#endif
    map->mapwin =
      XtCreatePopupShell("Map", topLevelShellWidgetClass,
			 toplevel(side), tmpargs, i);
    i = 0;
    mapform =
      XtCreateManagedWidget("MapMain", formWidgetClass,
			    map->mapwin, tmpargs, i);
    i = 0;
    mapcontrols =
      XtCreateManagedWidget("MapControls", boxWidgetClass,
			    mapform, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "<-");  ++i;
    tmpwidget =
      XtCreateManagedWidget("zoomIn", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, right_map, map);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "->");  ++i;
    tmpwidget =
      XtCreateManagedWidget("zoomIn", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, left_map, 1);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Up");  ++i;
    tmpwidget =
      XtCreateManagedWidget("zoomIn", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, up_map, map);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Down");  ++i;
    tmpwidget =
      XtCreateManagedWidget("zoomIn", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, down_map, map);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Zoom In");  ++i;
    tmpwidget =
      XtCreateManagedWidget("zoomIn", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, zoom_in_map, map);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Zoom Out");  ++i;
    tmpwidget =
      XtCreateManagedWidget("zoomIn", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, zoom_out_map, map);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Show Terrain");  ++i;
    tmpwidget =
      XtCreateManagedWidget("showTerrain", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, show_terrain_map, map);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Show Units");  ++i;
    tmpwidget =
      XtCreateManagedWidget("showUnits", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, show_units_map, map);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Show Names");  ++i;
    tmpwidget =
      XtCreateManagedWidget("showNames", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, show_names_map, map);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Show Grid");  ++i;
    tmpwidget =
      XtCreateManagedWidget("showHexGrid", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, show_grid_map, map);
    i = 0;
    tmpwidget =
      XtCreateManagedWidget("cancel", commandWidgetClass,
			    mapcontrols, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, cancel_map, map);
    i = 0;
    XtSetArg(tmpargs[i], XtNhorizDistance, 10);  ++i;
    XtSetArg(tmpargs[i], XtNfromHoriz, mapcontrols);  ++i;
    XtSetArg(tmpargs[i], XtNresizable, True);  ++i;
    XtSetArg(tmpargs[i], XtNwidth, 450);  ++i;
    XtSetArg(tmpargs[i], XtNheight, 350);  ++i;
    mapviewport =
      XtCreateManagedWidget("MapViewport", viewportWidgetClass,
			    mapform, tmpargs, i);
    i = 0;
/*    XtSetArg(tmpargs[i], XtNbackground, "black");  ++i; */
    XtSetArg(tmpargs[i], XtNwidth, 500);  ++i;
    XtSetArg(tmpargs[i], XtNheight, 500);  ++i;
    XtSetArg(tmpargs[i], XtNpower, 4);  ++i;
    XtSetArg(tmpargs[i], XtNside, side);  ++i;
    /* fails if we say just "createwidget", sigh */
    map->viewwin =
      XtCreateManagedWidget("Map View", mapWidgetClass,
			    mapviewport, tmpargs, i);
    XtPopup(map->mapwin, XtGrabNone);
    map->curtool = LOOKTOOL;
    map->curshifttool = UNITMOVETOOL;
#if 0
    map->controlb[LOOK].selected = TRUE;
    map->controlb[SHOWTERR].selected = map->drawterrain;
    map->controlb[SHOWPATS].selected = map->drawhexpats;
    map->controlb[SHOWEDGES].selected = map->drawhexedges;
    map->controlb[SHOWUNITS].selected = map->drawunits;
    map->controlb[SHOWNAMES].selected = map->drawnames;
    config_designer_tools(side, map);
#endif
#if 0
    /* Have the view itself be a retained window, cuts down on redraws */
    if ((retain = XGetDefault(side->ui->dpy, PROGRAMNAME, "Retain")) == NULL ||
	(*retain != 'N' && *retain != 'n')
	/* strcasecmp(retain, "no") */) {
	foo.backing_store = 1;
	XChangeWindowAttributes(side->ui->dpy, map->viewwin,
				CWBackingStore, &foo);
    }
#endif
    /* The first map is the "front" one. */
    side->ui->front = map;
    /* Remember one of the world maps as "the" world map. */
    if (isworld) {
	side->ui->world = map;
    }
}

void
open_closeup_from_list(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));
    List *list = (List *) clientdata;
    XawListReturnStruct *foo;

    foo = XawListShowCurrent(list->listwininfo);
    if (foo->list_index >= 0) {
	popup_unit_closeup(side,
			   find_unit_in_list(side, list, foo->list_index));
    } else {
	/* should have grayed out closeup button... */
    }
}

void
disband_from_list(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));
    List *list = (List *) clientdata;
    Unit *unit;
    XawListReturnStruct *foo;

    foo = XawListShowCurrent(list->listwininfo);
    if (foo->list_index >= 0) {
	unit = find_unit_in_list(side, list, foo->list_index);
	if (unit /* and permission */) {
	    do_disband_action(unit, unit);
	    /* when should this happen?? */
	    update_all_unit_displays(unit);
	}
    } else {
	/* should have grayed out disband button... */
    }
}

void
cancel_list(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));
    List *list = (List *) clientdata;

    /* should destroy list object too somehow */
    XtPopdown(list->listwin);
}

/* Create a window that lists units, unit types, etc. */

create_list(side, ltype)
Side *side;
int ltype;
{
    List *list = (List *) xmalloc(sizeof(List));
    int i, wid, hgt;
    Widget listbox, tmpwidget;

    list->type = ltype;
    list->next = side->ui->lists;
    side->ui->lists = list;

    list->listbuf = (char **) xmalloc (200 * sizeof(char *));
    for (i = 0; i < 200; ++i) {
	list->listbuf[i] = NULL;
    }
    list->listbuf[0] = "some units in some order";
    list->listbuf[1] = "second line";
    list->listlen = 2;

    list->listwin =
      XtCreatePopupShell("List", topLevelShellWidgetClass,
			 toplevel(side), NULL, 0);
    i = 0;
#if 0
    XtSetArg(tmpargs[i], XtNwidth, 600);  ++i;
    XtSetArg(tmpargs[i], XtNheight, 650);  ++i;
#endif
    listbox =
      XtCreateManagedWidget("box", boxWidgetClass,
			    list->listwin, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlist, list->listbuf);  ++i;
    XtSetArg(tmpargs[i], XtNnumberStrings, list->listlen);  ++i;
    XtSetArg(tmpargs[i], XtNverticalList, False); ++i;
    XtSetArg(tmpargs[i], XtNdefaultColumns, 1); ++i;
#if 0
    XtSetArg(tmpargs[i], XtNwidth, 300);  ++i;
    XtSetArg(tmpargs[i], XtNheight, 500);  ++i;
#endif
/* how to get the scroll bar? */
/*    XtSetArg(tmpargs[i], XtNtextOptions, scrollVertical);  ++i;  */
    list->listwininfo =
      XtCreateManagedWidget("ListContent", listWidgetClass,
			    listbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Closeup");  ++i;
    tmpwidget =
      XtCreateManagedWidget("closeup", commandWidgetClass,
			    listbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, open_closeup_from_list, list);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Disband");  ++i;
    tmpwidget =
      XtCreateManagedWidget("disband", commandWidgetClass,
			    listbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, disband_from_list, list);
    i = 0;
    tmpwidget =
      XtCreateManagedWidget("cancel", commandWidgetClass,
			    listbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, cancel_list, list);
    XtPopup(list->listwin, XtGrabNone);

    update_unit_list_content(side, list);

    side->ui->frontlist = list;
}

void
cancel_closeup(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));
    Closeup *closeup = (Closeup *) clientdata;

    /* should destroy closeup object too somehow */
    XtPopdown(closeup->closeupwin);
}

popup_unit_closeup(side, unit)
Side *side;
Unit *unit;
{
    int i;
    char *inside;
    Closeup *closeup = (Closeup *) xmalloc(sizeof(Closeup));
    Widget closeupbox, tmpwidget;

    if (unit == NULL) return;

    closeup->unit = unit;
    closeup->next = side->ui->closeups;
    side->ui->closeups = closeup;

    closeup->closeupwin =
      XtCreatePopupShell("Closeup", topLevelShellWidgetClass,
			 toplevel(side), NULL, 0);
    i = 0;
    closeupbox =
      XtCreateManagedWidget("box", boxWidgetClass,
			    closeup->closeupwin, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, unit_handle(side, unit));  ++i;
    tmpwidget =
      XtCreateManagedWidget("unit", labelWidgetClass,
			    closeupbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_container(side, unit));  ++i;
    tmpwidget =
      XtCreateManagedWidget("in", labelWidgetClass,
			    closeupbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_occupants(side, unit));  ++i;
    tmpwidget =
      XtCreateManagedWidget("occs", labelWidgetClass,
			    closeupbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_points(side, unit));  ++i;
    tmpwidget =
      XtCreateManagedWidget("points", labelWidgetClass,
			    closeupbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_supply(side, unit));  ++i;
    tmpwidget =
      XtCreateManagedWidget("supply", labelWidgetClass,
			    closeupbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_plan(side, unit));  ++i;
    tmpwidget =
      XtCreateManagedWidget("plan", labelWidgetClass,
			    closeupbox, tmpargs, i);
    i = 0;
    tmpwidget =
      XtCreateManagedWidget("cancel", commandWidgetClass,
			    closeupbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, cancel_closeup, closeup);
    XtPopup(closeup->closeupwin, XtGrabNone);

    side->ui->frontcloseup = closeup;
}

char *
describe_ask_side(side)
Side *side;
{
    return (side->unitseveraskside ? "ask" : "never ask");
}

void
ask_side_side_closeup(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i;
    Side *side = side_from_display(XtDisplay(w));
    SideCloseup *sidecloseup = (SideCloseup *) clientdata;

    side->unitseveraskside = !side->unitseveraskside;
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_ask_side(side));  ++i;
    XtSetValues(w, tmpargs, i);
}

char *
describe_use_machine_strategy(side)
Side *side;
{
    return (side->usemachinestrategy ? "machine" : "no machine");
}

void
use_machine_strategy_side_closeup(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i;
    Side *side = side_from_display(XtDisplay(w));
    SideCloseup *sidecloseup = (SideCloseup *) clientdata;

    side->usemachinestrategy = !side->usemachinestrategy;
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_use_machine_strategy(side));  ++i;
    XtSetValues(w, tmpargs, i);
}

char *
describe_auto_finish(side)
Side *side;
{
    return (side->autofinish ? "auto-finish turns" : "explicit finish");
}

void
auto_finish_side_closeup(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    int i;
    Side *side = side_from_display(XtDisplay(w));
    SideCloseup *sidecloseup = (SideCloseup *) clientdata;

    side->autofinish = !side->autofinish;
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_auto_finish(side));  ++i;
    XtSetValues(w, tmpargs, i);
}

char *
describe_turn_end_timeout(side)
Side *side;
{
    sprintf(spbuf, "Finish turn after %d seconds idle",
	    side->realtimeout);
    return spbuf;
}

void
turn_end_timeout_side_closeup(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
#if 0
    int i;
    Side *side = side_from_display(XtDisplay(w));
    SideCloseup *sidecloseup = (SideCloseup *) clientdata;

    side->unitseveraskside = !side->unitseveraskside;
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_ask_side(side));  ++i;
    XtSetValues(w, tmpargs, i);
#endif
}

void
cancel_side_closeup(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));
    SideCloseup *sidecloseup = (SideCloseup *) clientdata;

    /* should destroy sidecloseup object too somehow */
    XtPopdown(sidecloseup->closeupwin);
}

popup_side_closeup(side, side2)
Side *side, *side2;
{
    int i;
    SideCloseup *sidecloseup = (SideCloseup *) xmalloc(sizeof(SideCloseup));
    Widget closeupbox, doctrinebox, tmpwidget;

    if (side2 == NULL) return;

    sidecloseup->side = side2;
    sidecloseup->next = side->ui->sidecloseups;
    side->ui->sidecloseups = sidecloseup;

    sidecloseup->closeupwin =
      XtCreatePopupShell("SideCloseup", topLevelShellWidgetClass,
			 toplevel(side), NULL, 0);
    i = 0;
#if 0
    XtSetArg(tmpargs[i], XtNwidth, 600);  ++i;
    XtSetArg(tmpargs[i], XtNheight, 250);  ++i;
#endif
    closeupbox =
      XtCreateManagedWidget("box", boxWidgetClass,
			    sidecloseup->closeupwin, tmpargs, i);
    i = 0;
    if (0 /* side->type */) {
	sprintf(tmpbuf, "[%s]", "" /* side->type */);
    } else {
	sprintf(tmpbuf, "");
    }
    /* add username etc? */
    sprintf(spbuf, "%s%s", side2->name, tmpbuf);
    XtSetArg(tmpargs[i], XtNlabel, spbuf);  ++i;
    tmpwidget =
      XtCreateManagedWidget("SideName", labelWidgetClass,
			    closeupbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "relations");  ++i;
    tmpwidget =
      XtCreateManagedWidget("relations", labelWidgetClass,
			    closeupbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "tech");  ++i;
    tmpwidget =
      XtCreateManagedWidget("tech", labelWidgetClass,
			    closeupbox, tmpargs, i);
    /* Doctrine stuff all goes into a sub-box. */
    i = 0;
    doctrinebox =
      XtCreateManagedWidget("doctrine", boxWidgetClass,
			    closeupbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_ask_side(side));  ++i;
    tmpwidget =
      XtCreateManagedWidget("askSide", commandWidgetClass,
			    doctrinebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, ask_side_side_closeup, sidecloseup);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_use_machine_strategy(side));  ++i;
    tmpwidget =
      XtCreateManagedWidget("useMachineStrategy", commandWidgetClass,
			    doctrinebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, use_machine_strategy_side_closeup, sidecloseup);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_auto_finish(side));  ++i;
    tmpwidget =
      XtCreateManagedWidget("autoFinish", commandWidgetClass,
			    doctrinebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, auto_finish_side_closeup, sidecloseup);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, describe_turn_end_timeout(side));  ++i;
    tmpwidget =
      XtCreateManagedWidget("TimeOut", commandWidgetClass,
			    closeupbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, turn_end_timeout_side_closeup, sidecloseup);
    i = 0;
    tmpwidget =
      XtCreateManagedWidget("cancel", commandWidgetClass,
			    closeupbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, cancel_side_closeup, sidecloseup);
    XtPopup(sidecloseup->closeupwin, XtGrabNone);

    if (side == side2) {
	side->ui->ourcloseup = sidecloseup;
    }
}

/* Put in a default player, probably the invoker of the program. */

Player *
add_default_player()
{
    Player *player = add_player();

    player->name = getenv("USER");
    player->displayname = getenv("DISPLAY");

    return player;
}

void
side_closeup_game_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w)), *side2 = NULL;
    XawListReturnStruct *foo;

    foo = XawListShowCurrent(side->ui->sidenamelist);
    if (foo->list_index >= 0) {
	side2 = side_n(foo->list_index + 1);
    }
    if (side2 != NULL) {
	popup_side_closeup(side, side2);
    } else {
	beep(side);
    }
}

void
new_map_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    create_map(side, FALSE);
}

void
new_world_map_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    create_map(side, TRUE);
}

void
new_list_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    create_list(side, 1);
}

void
new_side_list_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    popup_sides_widget(side);
}

void
help_game_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    reveal_help(side);
}

void
save_game_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));
    Side *side2;

#if 0
    /* First check to see if anybody is in the middle of doing something
       (like renaming a unit) whose state would be lost. */
    for_all_sides(side2) {
	if (side2 != side && (side2->busy
			      || (side2->ui && side2->ui->modalhandler))) {
	    cmd_error("The %s are busy doing something, can't save!",
		      plural_form(side2->name));
	}
    }
#endif
    if (side->designer) {
	printf("should pop up a dialog");
    } else {
	save_the_game(side);
    }
}

void
resign_game_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    resign_game(side, NULL);
}

void
quit_game_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    /* Quit ungracefully */
    exit(0);
}

/* Explicit stepping command. */

void
step_game_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    run_game(10);
}

void
debug_game_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    switch ((int) calldata) {
      case 0:
	Debug = !Debug;
	return;
      case 1:
	DebugM = !DebugM;
	return;
      case 2:
	DebugG = !DebugG;
	return;
    }
}

void
design_game_command(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    /* should check password eventually? */
    if (side->designer) --numdesigners; else ++numdesigners;
    side->designer = !side->designer;
    config_designer_tools(side, NULL);
}

popup_game_widget(side)
Side *side;
{
    int i;
    Widget gamebox, tmpwidget;

    side->ui->game =
      XtCreatePopupShell("RandomCommands", topLevelShellWidgetClass,
			 toplevel(side), NULL, 0);
    i = 0;
    XtSetArg(tmpargs[i], XtNwidth, 300);  ++i;
    XtSetArg(tmpargs[i], XtNheight, 200);  ++i;
    gamebox =
      XtCreateManagedWidget("Box", boxWidgetClass, side->ui->game, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNwidth, 200);  ++i;
    XtSetArg(tmpargs[i], XtNlabel, "Turn 0");  ++i;
    side->ui->datelabel =
      XtCreateManagedWidget("Date", labelWidgetClass,
			    gamebox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "New Map");  ++i;
    tmpwidget =
      XtCreateManagedWidget("NewMap", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, new_map_command, NULL);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "New World Map");  ++i;
    tmpwidget =
      XtCreateManagedWidget("NewWorldMap", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, new_world_map_command, NULL);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "New List");  ++i;
    tmpwidget =
      XtCreateManagedWidget("NewList", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, new_list_command, NULL);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "New Side List");  ++i;
    tmpwidget =
      XtCreateManagedWidget("NewSideList", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, new_side_list_command, NULL);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Help");  ++i;
    tmpwidget =
      XtCreateManagedWidget("Help", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, help_game_command, NULL);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Save");  ++i;
    tmpwidget =
      XtCreateManagedWidget("Save", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, save_game_command, NULL);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Resign");  ++i;
    tmpwidget =
      XtCreateManagedWidget("Resign", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, resign_game_command, NULL);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "QUIT!");  ++i;
    tmpwidget =
      XtCreateManagedWidget("Quit", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, quit_game_command, NULL);
    /* Debugging group. */
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Step");  ++i;
    tmpwidget =
      XtCreateManagedWidget("Step", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, step_game_command, NULL);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Design");  ++i;
    tmpwidget =
      XtCreateManagedWidget("Design", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, design_game_command, 0);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "D");  ++i;
    tmpwidget =
      XtCreateManagedWidget("D", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, debug_game_command, 0);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "DM");  ++i;
    tmpwidget =
      XtCreateManagedWidget("DM", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, debug_game_command, 1);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "DG");  ++i;
    tmpwidget =
      XtCreateManagedWidget("DG", commandWidgetClass,
			    gamebox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, debug_game_command, 2);
    XtPopup(side->ui->game, XtGrabNone);
}

void
side_closeup_sides_list(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w)), *side2 = NULL;
    XawListReturnStruct *foo;

    foo = XawListShowCurrent(side->ui->sidenamelist);
    if (foo->list_index >= 0) {
	side2 = side_n(foo->list_index + 1);
    }
    if (side2 != NULL) {
	popup_side_closeup(side, side2);
    } else {
	beep(side);
    }
}

void
cancel_sides_list(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    /* should destroy sidecloseup object too somehow */
    XtPopdown(side->ui->sides);
}

popup_sides_widget(side)
Side *side;
{
    int i;
    Widget sidesbox, tmpwidget;

    i = 0;
#if 0
    XtSetArg(tmpargs[i], XtNx, 400);  ++i;
    XtSetArg(tmpargs[i], XtNy, 50);  ++i;
    XtSetArg(tmpargs[i], XtNwidth, 300);  ++i;
    XtSetArg(tmpargs[i], XtNheight, 200);  ++i;
#endif
    side->ui->sides =
      XtCreatePopupShell("SidesSides", topLevelShellWidgetClass,
			 toplevel(side), tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNwidth, 300);  ++i;
    XtSetArg(tmpargs[i], XtNheight, 200);  ++i;
    sidesbox =
      XtCreateManagedWidget("Box", boxWidgetClass,
			    side->ui->sides, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNwidth, 300);  ++i;
    XtSetArg(tmpargs[i], XtNheight, 200);  ++i;
    {
	int j = 0;
	Side *side2;

	for_all_sides(side2) {
	    side->ui->sidestrings[j++] = 
	      copy_string(describe_side_state(side, side2));
	}
    }
    XtSetArg(tmpargs[i], XtNlist, side->ui->sidestrings);  ++i;
    XtSetArg(tmpargs[i], XtNnumberStrings, numsides);  ++i;
    XtSetArg(tmpargs[i], XtNverticalList, True); ++i;
    XtSetArg(tmpargs[i], XtNdefaultColumns, 1); ++i;
    side->ui->sidenamelist =
      XtCreateManagedWidget("SideList", listWidgetClass,
			    sidesbox, tmpargs, i);
    i = 0;
    XtSetArg(tmpargs[i], XtNlabel, "Side Closeup");  ++i;
    tmpwidget =
      XtCreateManagedWidget("SideCloseup", commandWidgetClass,
			    sidesbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, side_closeup_sides_list, NULL);
    i = 0;
    tmpwidget =
      XtCreateManagedWidget("Cancel", commandWidgetClass,
			    sidesbox, tmpargs, i);
    XtAddCallback(tmpwidget, XtNcallback, cancel_sides_list, NULL);

    /* should desensitive side list opener? */

    XtPopup(side->ui->sides, XtGrabNone);
}

/* A predicate that tests whether our display can safely be written to. */

active_display(side)
Side *side;
{
    return (side && ((int) side->ui) > 1 && side->ui->active);
}

set_side_list_content(side, list, str)
Side *side;
List *list;
char *str;
{
    int i = 0;

    XtSetArg(tmpargs[i], XtNstring, str);  ++i;
    XtSetValues(list->listwininfo, tmpargs, i);

    XawTextSetInsertionPoint(list->listwininfo, 0);
}

/* Compute the distinctive emblem by which one side will recognize units
   of another side.  This does both our view of ourselves and others
   orthogonally.  Note that sides without displays can still have emblems
   and colors that the sides with displays will see. */

init_emblem(side, side2)
Side *side, *side2;
{
    int s2 = side_number(side2), i;
    int w, h;

    /* Collect the color names of the other side and try to request
       them for our own display. */
    side->ui->numcolors[s2] = (side2->colorscheme ? 1 : 0);
    /* eventually take spec apart and iterate for multiple colors */
    side->ui->colors[s2][0] = request_color(side, side2->colorscheme);
    /* Turn emblem names into bitmaps. */
    if (side2->emblemname) {
	side->ui->emblempics[s2] =
	  load_bitmap(side, side2->emblemname, "e", &w, &h);
	/* necessary? */
	if (side->ui->emblempics[s2] != None) {
	    side->ui->emblems[s2] = TRUE;
	    side->ui->embw[s2] = w;  side->ui->embh[s2] = h;
	}
    } else {
	/* Still need a width and height. */
    }
}

/* These return emblem dimensions so higher-level code can make good
   formatting decisions. */

emblem_width(side, side2)
Side *side, *side2;
{
    return side->ui->embw[side_number(side2)]; 
}

emblem_height(side, side2)
Side *side, *side2;
{
    return side->ui->embh[side_number(side2)]; 
}

/* Enable/disable the designer's control panel (on one map or all). */

config_designer_tools(side, map)
Side *side;
Map *map;
{
    int i;

    if (map == NULL) {
	for_all_maps(side, map) {
	    config_designer_tools_for_map(side, map);
/*	    draw_controls(side, map);  */
	}
    } else {
	config_designer_tools_for_map(side, map);
    }
}

/* Set/reset the states of the designer controls. */

/* (this should be ifdefed for linear features, etc) */

config_designer_tools_for_map(side, map)
Side *side;
Map *map;
{
#if 0
    int i;

    for (i = 0; i < NUMCONTROLS; ++i) {
	/* If a selected tool is designer-only, switch to another. */
	if (!side->designer
	    && controltypes[i].fordesigners
	    && controltypes[i].tool >= 0
	    && map->controlb[i].selected) {
	    map->controlb[i].selected = TRUE;
	    map->controlb[LOOK].selected = TRUE;
	}
	if (controltypes[i].fordesigners) {
	    map->controlb[i].visible = side->designer;
	    map->controlb[i].enabled = side->designer;
	    /* Selection status should not be affected by default. */
	}
    }
    /* Make sure these have something reasonable in them. */
    map->controlb[CURTTYPE].color =
      side->ui->hexcolor[map->curttype];
    map->controlb[CURUTYPE].pic =
      side->ui->unitpics[4][map->curutype];
    if (map->controlb[CURSIDE].text == NULL) {
	sprintf(tmpbuf, "%d", side_number(side));
	map->controlb[CURSIDE].text = copy_string(tmpbuf);
    }
    /* curside's color/pic should be side's color/emblem */
#endif
}

/* Shut the side's display down and flag it as having been done. */

close_display(side)
Side *side;
{
    XtCloseDisplay(side->ui->dpy);
    side->ui->active = FALSE;
}
